import json
import os
from datetime import date
from unittest import TestCase

from basketball_reference_web_scraper.client import players_season_totals
from basketball_reference_web_scraper.data import OutputType, OutputWriteOption
from basketball_reference_web_scraper.errors import InvalidSeason


class BaseTestPlayerSeasonTotalsJSONFileOutput(TestCase):
    @property
    def year(self):
        raise NotImplementedError

    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/player_season_totals_{year}.json".format(year=self.year),
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/player_season_totals_{year}.json".format(year=self.year)
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def assert_player_season_totals_file(self):
        players_season_totals(
            season_end_year=self.year,
            output_type=OutputType.JSON,
            output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )
        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            self.assertEqual(
                json.load(output_file),
                json.load(expected_output_file),
            )


class BaseTestPlayerSeasonTotalsCSVFileOutput(TestCase):
    @property
    def year(self):
        raise NotImplementedError

    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/player_season_totals_{year}.csv".format(year=self.year),
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/player_season_totals_{year}.csv".format(year=self.year)
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def assert_player_season_totals_file(self):
        players_season_totals(
            season_end_year=self.year,
            output_type=OutputType.CSV,
            output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )
        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            self.assertEqual(
                output_file.readlines(),
                expected_output_file.readlines(),
            )


class TestPlayerSeasonTotalsInMemory(TestCase):
    def test_2018_players_season_totals_length(self):
        result = players_season_totals(season_end_year=2018)
        self.assertEqual(len(result), 605)

    def test_future_season_raises_invalid_season(self):
        current_year = date.today().year
        future_year = current_year + 10
        expected_message = "Season end year of {future_year} is invalid".format(future_year=future_year)
        self.assertRaisesRegex(InvalidSeason, expected_message, players_season_totals, season_end_year=future_year)


class Test2018PlayerSeasonTotalsInMemoryJSONOutput(TestCase):
    def setUp(self):
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/player_season_totals_2018.json"
        )

    def test_players_season_totals_json(self):
        result = players_season_totals(season_end_year=2018, output_type=OutputType.JSON)
        with open(self.expected_output_file_path, "r") as expected_results_file:
            self.assertEqual(
                json.loads(result),
                json.load(expected_results_file)
            )


class Test2018PlayerSeasonTotalsCSVFileOutput(BaseTestPlayerSeasonTotalsCSVFileOutput):
    @property
    def year(self):
        return 2018

    def test_2018_player_season_totals_file(self):
        self.assert_player_season_totals_file()


class Test2001PlayerSeasonTotalsCSVFileOutput(BaseTestPlayerSeasonTotalsCSVFileOutput):
    @property
    def year(self):
        return 2001

    def test_2001_player_season_totals_file(self):
        self.assert_player_season_totals_file()


class Test2018PlayerSeasonTotalsJSONFileOutput(BaseTestPlayerSeasonTotalsJSONFileOutput):
    @property
    def year(self):
        return 2018

    def test_2018_player_season_totals_file(self):
        self.assert_player_season_totals_file()




from unittest import TestCase

from basketball_reference_web_scraper.data import Outcome, OUTCOME_ABBREVIATIONS_TO_OUTCOME
from basketball_reference_web_scraper.parsers import OutcomeAbbreviationParser


class TestOutcomeAbbreviationParser(TestCase):
    def setUp(self):
        self.parser = OutcomeAbbreviationParser(abbreviations_to_outcomes=OUTCOME_ABBREVIATIONS_TO_OUTCOME)

    def test_parse_unknown_outcome_symbol(self):
        self.assertRaises(ValueError, self.parser.from_abbreviation, "jaebaebae")

    def test_parse_win(self):
        self.assertEqual(Outcome.WIN, self.parser.from_abbreviation("W"))

    def test_parse_loss(self):
        self.assertEqual(Outcome.LOSS, self.parser.from_abbreviation("L"))



def str_to_int(value, default=int(0)):
    stripped_value = value.strip()
    try:
        return int(stripped_value)
    except ValueError:
        return default


def str_to_float(value, default=float(0)):
    stripped_value = value.strip()
    try:
        return float(stripped_value)
    except ValueError:
        return default


def merge_two_dicts(first, second):
    combined = first.copy()
    combined.update(second)
    return combined



from unittest import TestCase
from unittest.mock import patch, MagicMock

from requests import HTTPError, codes

from basketball_reference_web_scraper.client import players_advanced_season_totals
from basketball_reference_web_scraper.errors import InvalidSeason


class TestPlayerAdvancedSeasonTotals(TestCase):
    @patch("basketball_reference_web_scraper.http_client.players_advanced_season_totals")
    def test_not_found_raises_invalid_season(self, mocked_players_advanced_season_totals):
        end_year = "jaebaebae"
        expected_message = "Season end year of {end_year} is invalid".format(end_year=end_year)
        mocked_players_advanced_season_totals.side_effect = HTTPError(
            response=MagicMock(status_code=codes.not_found))
        self.assertRaisesRegex(InvalidSeason, expected_message, players_advanced_season_totals, season_end_year=end_year)

    @patch("basketball_reference_web_scraper.http_client.players_advanced_season_totals")
    def test_other_http_error_is_raised(self, mocked_players_advanced_season_totals):
        mocked_players_advanced_season_totals.side_effect = HTTPError(
            response=MagicMock(status_code=codes.internal_server_error))
        self.assertRaises(HTTPError, players_advanced_season_totals, season_end_year=2018)



from unittest import TestCase

from basketball_reference_web_scraper.parsers import SearchResultNameParser


class TestSearchResultNameParser(TestCase):
    def setUp(self):
        self.parser = SearchResultNameParser()

    def test_parse_name_with_parentheses_with_start_and_end_year(self):
        self.assertEqual(
            self.parser.parse(search_result_name='Kobe Bryant (1997-2016)'),
            'Kobe Bryant',
        )

    def test_parse_name_with_parentheses_with_start_year(self):
        self.assertEqual(
            self.parser.parse(search_result_name='Bud Koper (1965)'),
            'Bud Koper',
        )

    def test_parse_name_without_parentheses(self):
        self.assertEqual(
            self.parser.parse(search_result_name='Bronson Koenig'),
            'Bronson Koenig',
        )




from unittest import TestCase, mock

from requests import HTTPError, codes

from basketball_reference_web_scraper.client import play_by_play
from basketball_reference_web_scraper.data import Team
from basketball_reference_web_scraper.errors import InvalidDate


class TestPlayByPlay(TestCase):
    @mock.patch("basketball_reference_web_scraper.client.http_client")
    def test_raises_invalid_date_for_404_response(self, mocked_http_client):
        mocked_http_client.play_by_play.side_effect = HTTPError(response=mock.Mock(status_code=codes.not_found))
        self.assertRaises(InvalidDate, play_by_play, home_team=Team.MILWAUKEE_BUCKS,  day=1, month=1, year=2018)

    @mock.patch("basketball_reference_web_scraper.client.http_client")
    def test_raises_non_404_http_error(self, mocked_http_client):
        mocked_http_client.play_by_play.side_effect = HTTPError(response=mock.Mock(status_code=codes.server_error))
        self.assertRaises(HTTPError, play_by_play, home_team=Team.MILWAUKEE_BUCKS,  day=1, month=1, year=2018)




from unittest import TestCase
from unittest.mock import MagicMock, patch, PropertyMock

from basketball_reference_web_scraper.html import PlayerSeasonTotalsRow


class TestPlayerSeasonTotalsRow(TestCase):
    def setUp(self):
        self.html = MagicMock()

    def test_position_abbreviations_when_cells_exist(self):
        self.html.xpath = MagicMock(
            return_value=[
                MagicMock(text_content=MagicMock(return_value="some position abbreviations"))
            ]
        )

        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).position_abbreviations, "some position abbreviations")
        self.html.xpath.assert_called_once_with('td[@data-stat="pos"]')

    def test_position_abbreviations_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).position_abbreviations, "")
        self.html.xpath.assert_called_once_with('td[@data-stat="pos"]')

    def test_age_when_cells_exist(self):
        self.html.xpath = MagicMock(
            return_value=[
                MagicMock(text_content=MagicMock(return_value="some age"))
            ]
        )

        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).age, "some age")
        self.html.xpath.assert_called_once_with('td[@data-stat="age"]')

    def test_age_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).age, "")
        self.html.xpath.assert_called_once_with('td[@data-stat="age"]')

    def test_games_played_when_cells_exist(self):
        self.html.xpath = MagicMock(
            return_value=[
                MagicMock(text_content=MagicMock(return_value="some games played"))
            ]
        )

        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).games_played, "some games played")
        self.html.xpath.assert_called_once_with('td[@data-stat="g"]')

    def test_games_played_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).games_played, "")
        self.html.xpath.assert_called_once_with('td[@data-stat="g"]')

    def test_games_started_when_cells_exist(self):
        self.html.xpath = MagicMock(
            return_value=[
                MagicMock(text_content=MagicMock(return_value="some games started"))
            ]
        )

        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).games_started, "some games started")
        self.html.xpath.assert_called_once_with('td[@data-stat="gs"]')

    def test_games_started_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerSeasonTotalsRow(html=self.html).games_started, "")
        self.html.xpath.assert_called_once_with('td[@data-stat="gs"]')

    @patch.object(PlayerSeasonTotalsRow, 'team_abbreviation', new_callable=PropertyMock)
    def test_is_combined_totals_when_team_abbreviation_is_tot(self, mocked_team_abbreviation):
        mocked_team_abbreviation.return_value = "TOT"
        self.assertTrue(PlayerSeasonTotalsRow(html=self.html).is_combined_totals)

    @patch.object(PlayerSeasonTotalsRow, 'team_abbreviation', new_callable=PropertyMock)
    def test_is_not_combined_totals_when_team_abbreviation_is_tot(self, mocked_team_abbreviation):
        mocked_team_abbreviation.return_value = "jaebaebae"
        self.assertFalse(PlayerSeasonTotalsRow(html=self.html).is_combined_totals)



from unittest import TestCase
from unittest.mock import MagicMock, PropertyMock, patch

from basketball_reference_web_scraper.html import SearchResult


class TestSearchResult(TestCase):
    def test_resource_link_query(self):
        self.assertEqual(
            SearchResult(html=MagicMock()).resource_link_query,
            './div[@class="search-item-name"]//a',
        )

    @patch.object(SearchResult, "resource_link_query", new_callable=PropertyMock)
    def test_resource_link_when_no_matching_links(self, mocked_query):
        mocked_query.return_value = "some query"
        html = MagicMock()
        html.xpath = MagicMock(return_value=[])

        self.assertIsNone(SearchResult(html=html).resource_link)
        html.xpath.assert_called_once_with("some query")

    @patch.object(SearchResult, "resource_link_query", new_callable=PropertyMock)
    def test_resource_link_when_matching_links(self, mocked_query):
        mocked_query.return_value = "some query"
        first_link = MagicMock()
        html = MagicMock()
        html.xpath = MagicMock(return_value=[first_link])

        self.assertEqual(
            SearchResult(html=html).resource_link,
            first_link,
        )
        html.xpath.assert_called_once_with("some query")

    @patch.object(SearchResult, "resource_link", new_callable=PropertyMock)
    def test_resource_location_when_resource_link_is_none(self, mocked_resource_link):
        mocked_resource_link.return_value = None
        self.assertIsNone(SearchResult(html=MagicMock()).resource_location)

    @patch.object(SearchResult, "resource_link", new_callable=PropertyMock)
    def test_resource_location_when_resource_link_is_not_none(self, mocked_resource_link):
        link = MagicMock()
        link.attrib = MagicMock()
        link.attrib.__getitem__ = MagicMock(return_value="some href")
        mocked_resource_link.return_value = link

        self.assertEqual(
            SearchResult(html=MagicMock()).resource_location,
            "some href"
        )
        link.attrib.__getitem__.assert_called_once_with("href")

    @patch.object(SearchResult, "resource_link", new_callable=PropertyMock)
    def test_resource_name_when_resource_link_is_none(self, mocked_resource_link):
        mocked_resource_link.return_value = None
        self.assertIsNone(SearchResult(html=MagicMock()).resource_name)

    @patch.object(SearchResult, "resource_link", new_callable=PropertyMock)
    def test_resource_name_when_resource_link_is_not_none(self, mocked_resource_link):
        link = MagicMock()
        link.text_content = MagicMock(return_value="some content")
        mocked_resource_link.return_value = link
        self.assertEqual(
            SearchResult(html=MagicMock()).resource_name,
            "some content"
        )
        link.text_content.assert_called_once_with()

    def test_different_class_is_not_equal(self):
        self.assertNotEqual(
            SearchResult(html=MagicMock()),
            "jaebaebae"
        )

    def test_different_html_but_same_class_is_not_equal(self):
        self.assertNotEqual(
            SearchResult(html=MagicMock()),
            SearchResult(html=MagicMock())
        )

    def test_same_html_and_same_class_is_equal(self):
        html = MagicMock()
        self.assertEqual(
            SearchResult(html=html),
            SearchResult(html=html),
        )




from unittest import TestCase
from unittest.mock import MagicMock

from basketball_reference_web_scraper.html import RegularSeasonPlayerBoxScoresTable, PlayerSeasonBoxScoresRow


class TestRegularSeasonPlayerBoxScoresTable(TestCase):
    def setUp(self):
        self.html = MagicMock()

    def test_rows_query(self):
        self.assertEqual(
            '//tbody/tr[not(contains(@class, "thead"))]',
            RegularSeasonPlayerBoxScoresTable(html=self.html).rows_query,
        )

    def test_rows_returns_empty_array_when_there_are_not_any_matching_rows(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertListEqual([], RegularSeasonPlayerBoxScoresTable(html=self.html).rows)

    def test_rows_returns_populated_array_when_there_are_matching_rows(self):
        first_row_html = MagicMock(name="first matching row html")
        second_row_html = MagicMock(name="second matching row html")
        self.html.xpath = MagicMock(return_value=[first_row_html, second_row_html])
        self.assertListEqual([
                PlayerSeasonBoxScoresRow(html=first_row_html),
                PlayerSeasonBoxScoresRow(html=second_row_html),
            ],
            RegularSeasonPlayerBoxScoresTable(html=self.html).rows,
        )



from unittest import TestCase

from basketball_reference_web_scraper.data import OUTCOME_ABBREVIATIONS_TO_OUTCOME, Outcome
from basketball_reference_web_scraper.parsers import OutcomeAbbreviationParser, PlayerBoxScoreOutcomeParser


class TestPlayerBoxScoreOutcomeParser(TestCase):
    def setUp(self):
        self.parser = PlayerBoxScoreOutcomeParser(
            outcome_abbreviation_parser=OutcomeAbbreviationParser(
                abbreviations_to_outcomes=OUTCOME_ABBREVIATIONS_TO_OUTCOME
            )
        )

    def test_parse_win_abbreviation_for_single_digit_margin_of_victory(self):
        self.assertEqual("W", self.parser.parse_outcome_abbreviation(formatted_outcome="W (+8)"))

    def test_parse_win_abbreviation_for_double_digit_margin_of_victory(self):
        self.assertEqual("W", self.parser.parse_outcome_abbreviation(formatted_outcome="W (+18)"))

    def test_parse_loss_abbreviation_for_single_digit_margin_of_victory(self):
        self.assertEqual("L", self.parser.parse_outcome_abbreviation(formatted_outcome="L (-8)"))

    def test_parse_loss_abbreviation_for_double_digit_margin_of_victory(self):
        self.assertEqual("L", self.parser.parse_outcome_abbreviation(formatted_outcome="L (-18)"))

    def test_parse_win_outcome_for_single_digit_margin_of_victory(self):
        self.assertEqual(Outcome.WIN, self.parser.parse_outcome(formatted_outcome="W (+8)"))

    def test_parse_win_outcome_for_double_digit_margin_of_victory(self):
        self.assertEqual(Outcome.WIN, self.parser.parse_outcome(formatted_outcome="W (+18)"))

    def test_parse_loss_outcome_for_single_digit_margin_of_victory(self):
        self.assertEqual(Outcome.LOSS, self.parser.parse_outcome(formatted_outcome="L (-8)"))

    def test_parse_loss_outcome_for_double_digit_margin_of_victory(self):
        self.assertEqual(Outcome.LOSS, self.parser.parse_outcome(formatted_outcome="L (-18)"))

    def test_parse_positive_single_digit_margin_of_victory(self):
        self.assertEqual(8, self.parser.parse_margin_of_victory(formatted_outcome="W (+8)"))

    def test_parse_positive_double_digit_margin_of_victory(self):
        self.assertEqual(18, self.parser.parse_margin_of_victory(formatted_outcome="W (+18)"))

    def test_parse_negative_single_digit_margin_of_victory(self):
        self.assertEqual(-8, self.parser.parse_margin_of_victory(formatted_outcome="L (-8)"))

    def test_parse_negative_double_digit_margin_of_victory(self):
        self.assertEqual(-18, self.parser.parse_margin_of_victory(formatted_outcome="L (-18)"))



import json
import os
from unittest import TestCase

from basketball_reference_web_scraper.client import play_by_play
from basketball_reference_web_scraper.data import OutputType, OutputWriteOption, PeriodType
from basketball_reference_web_scraper.data import Team
from basketball_reference_web_scraper.errors import InvalidDate


class TestPlayByPlayInMemory(TestCase):
    def test_total_play_by_play_length_for_2018_10_29(self):
        result = play_by_play(home_team=Team.MILWAUKEE_BUCKS, day=27, month=10, year=2018)
        self.assertEqual(len(result), 465)

    def test_first_play_by_play_for_2018_10_29(self):
        result = play_by_play(home_team=Team.MILWAUKEE_BUCKS, day=27, month=10, year=2018)
        self.assertEqual(
            result[0],
            {
                "period": 1,
                "period_type": PeriodType.QUARTER,
                "relevant_team": Team.ORLANDO_MAGIC,
                "away_team": Team.ORLANDO_MAGIC,
                "home_team": Team.MILWAUKEE_BUCKS,
                "away_score": 0,
                "home_score": 0,
                "description": "N. Vučević misses 2-pt hook shot from 3 ft",
                "remaining_seconds_in_period": 703.0,
            }
        )

    def test_last_play_by_play_for_2018_10_29(self):
        result = play_by_play(home_team=Team.MILWAUKEE_BUCKS, day=27, month=10, year=2018)
        self.assertEqual(
            result[464],
            {
                "period": 4,
                "period_type": PeriodType.QUARTER,
                "relevant_team": Team.MILWAUKEE_BUCKS,
                "away_team": Team.ORLANDO_MAGIC,
                "home_team": Team.MILWAUKEE_BUCKS,
                "away_score": 91,
                "home_score": 113,
                "description": "Defensive rebound by T. Maker",
                "remaining_seconds_in_period": 2.0,
            }
        )

    def test_total_play_by_play_length_for_single_digit_month_and_day(self):
        result = play_by_play(home_team=Team.DENVER_NUGGETS, day=1, month=1, year=2019)
        self.assertEqual(len(result), 464)

    def test_first_play_by_play_for_2019_01_01(self):
        result = play_by_play(home_team=Team.DENVER_NUGGETS, day=1, month=1, year=2019)
        self.assertEqual(
            result[0],
            {
                "period": 1,
                "period_type": PeriodType.QUARTER,
                "relevant_team": Team.DENVER_NUGGETS,
                "away_team": Team.NEW_YORK_KNICKS,
                "home_team": Team.DENVER_NUGGETS,
                "away_score": 0,
                "home_score": 0,
                "description": "M. Plumlee misses 2-pt hook shot from 5 ft",
                "remaining_seconds_in_period": 693.0,
            }
        )

    def test_last_play_by_play_for_2019_01_01(self):
        result = play_by_play(home_team=Team.DENVER_NUGGETS, day=1, month=1, year=2019)
        self.assertEqual(
            result[463],
            {
                "period": 4,
                "period_type": PeriodType.QUARTER,
                "relevant_team": Team.DENVER_NUGGETS,
                "away_team": Team.NEW_YORK_KNICKS,
                "home_team": Team.DENVER_NUGGETS,
                "away_score": 108,
                "home_score": 115,
                "description": "Defensive rebound by M. Beasley",
                "remaining_seconds_in_period": 12.0,
            }
        )

    def test_last_play_by_play_for_overtime_game(self):
        result = play_by_play(home_team=Team.SACRAMENTO_KINGS, day=1, month=1, year=2019)
        self.assertEqual(
            result[507],
            {
                "period": 1,
                "period_type": PeriodType.OVERTIME,
                "relevant_team": Team.PORTLAND_TRAIL_BLAZERS,
                "away_team": Team.PORTLAND_TRAIL_BLAZERS,
                "home_team": Team.SACRAMENTO_KINGS,
                "away_score": 113,
                "home_score": 108,
                "description": "Defensive rebound by J. Nurkić",
                "remaining_seconds_in_period": 3.0,
            }
        )

    def test_get_play_by_play_for_day_that_does_not_exist(self):
        self.assertRaisesRegex(
            InvalidDate,
            "Date with year set to 2018, month set to 1, and day set to -1 is invalid",
            play_by_play,
            home_team=Team.MILWAUKEE_BUCKS, day=-1, month=1, year=2018)

    def test_raises_invalid_date_for_nonexistent_dates(self):
        self.assertRaisesRegex(
            InvalidDate,
            "Date with year set to baz, month set to bar, and day set to foo is invalid",
            play_by_play,
            home_team=Team.MILWAUKEE_BUCKS, day="foo", month="bar", year="baz")

    def test_non_unicode_matches(self):
        result = play_by_play(home_team=Team.GOLDEN_STATE_WARRIORS, day=16, month=10, year=2018)
        self.assertIsNotNone(result)


class TestPlayByPlayCSVOutput(TestCase):
    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2003_10_29_TOR_pbp.csv",
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2003_10_29_TOR_pbp.csv"
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def test_2003_10_29_TOR_box_scores(self):
        play_by_play(
            home_team=Team.TORONTO_RAPTORS, day=29, month=10, year=2003,
            output_type=OutputType.CSV,
            output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )
        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            self.assertEqual(output_file.readlines(), expected_output_file.readlines())


class TestPlayByPlayJSONOutput(TestCase):
    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2003_10_29_TOR_pbp.json",
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2003_10_29_TOR_pbp.json",
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def test_get_box_scores_from_2003_json(self):
        play_by_play(
            home_team=Team.TORONTO_RAPTORS, day=29, month=10, year=2003,
            output_type=OutputType.JSON, output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )

        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            self.assertEqual(
                json.load(output_file),
                json.load(expected_output_file),
            )



from unittest import TestCase, mock

from basketball_reference_web_scraper.data import OutputWriteOption
from basketball_reference_web_scraper.writers import CSVWriter, WriteOptions


class TestCSVWriter(TestCase):
    DATA = ["some", "row", "data"]
    COLUMN_NAMES = ["some", "column", "names"]

    @mock.patch("basketball_reference_web_scraper.writers.csv.DictWriter")
    def test_opens_correct_file(self, mock_csv_dict_writer):
        with mock.patch("builtins.open", mock.mock_open()) as mock_file:
            csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
            mock_csv_dict_writer.return_value = csv_dict_writer
            row_formatter = mock.Mock(format=mock.Mock())
            csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
            mock_csv_dict_writer.return_value = csv_dict_writer
            writer = CSVWriter(column_names=self.COLUMN_NAMES, row_formatter=row_formatter)
            writer.write(
                data=self.DATA,
                options=WriteOptions(
                    file_path="some file path",
                    mode=OutputWriteOption.WRITE
                )
            )
            mock_file.assert_called_with("some file path", OutputWriteOption.WRITE.value, newline="")

    @mock.patch("basketball_reference_web_scraper.writers.csv.DictWriter")
    def test_file_and_columns_are_used_by_writer(self, mock_csv_dict_writer):
        with mock.patch("builtins.open", mock.mock_open()) as mock_file:
            csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
            mock_csv_dict_writer.return_value = csv_dict_writer
            row_formatter = mock.Mock(format=mock.Mock())
            csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
            mock_csv_dict_writer.return_value = csv_dict_writer
            writer = CSVWriter(column_names=self.COLUMN_NAMES, row_formatter=row_formatter)
            writer.write(
                data=self.DATA,
                options=WriteOptions(
                    file_path="some file path",
                    mode=OutputWriteOption.WRITE
                )
            )
            mock_csv_dict_writer.assert_called_with(mock_file(), fieldnames=self.COLUMN_NAMES)

    @mock.patch("basketball_reference_web_scraper.writers.csv.DictWriter")
    def test_header_is_written(self, mock_csv_dict_writer):
        csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
        mock_csv_dict_writer.return_value = csv_dict_writer
        row_formatter = mock.Mock(format=mock.Mock())
        csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
        mock_csv_dict_writer.return_value = csv_dict_writer
        writer = CSVWriter(column_names=self.COLUMN_NAMES, row_formatter=row_formatter)
        writer.write(
            data=self.DATA,
            options=WriteOptions(
                file_path="some file path",
                mode=OutputWriteOption.WRITE
            )
        )
        csv_dict_writer.writeheader.assert_called_once_with()

    @mock.patch("basketball_reference_web_scraper.writers.csv.DictWriter")
    def test_rows_are_written(self, mock_csv_dict_writer):
        csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
        mock_csv_dict_writer.return_value = csv_dict_writer
        row_formatter = mock.Mock(format=mock.Mock())
        csv_dict_writer = mock.Mock(wrteheader=mock.Mock(), writerows=mock.Mock())
        mock_csv_dict_writer.return_value = csv_dict_writer
        writer = CSVWriter(column_names=self.COLUMN_NAMES, row_formatter=row_formatter)
        writer.write(
            data=self.DATA,
            options=WriteOptions(
                file_path="some file path",
                mode=OutputWriteOption.WRITE
            )
        )
        csv_dict_writer.writerows.assert_called_once_with(mock.ANY)
        self.assertEqual(3, row_formatter.format.call_count)
        row_formatter.format.assert_has_calls(
            calls=[
                mock.call("some"),
                mock.call("row"),
                mock.call("data"),
            ],
            any_order=False
        )



from unittest import TestCase

from basketball_reference_web_scraper.data import League, LEAGUE_ABBREVIATIONS_TO_LEAGUE
from basketball_reference_web_scraper.parsers import LeagueAbbreviationParser


class TestLeagueAbbreviationParser(TestCase):
    def setUp(self):
        self.parser = LeagueAbbreviationParser(abbreviations_to_league=LEAGUE_ABBREVIATIONS_TO_LEAGUE)

    def test_parsing_unknown_abbreviation(self):
        self.assertRaisesRegex(
            ValueError,
            "Unknown league abbreviation: jaebaebae",
            self.parser.from_abbreviation,
            abbreviation="jaebaebae",
        )

    def test_parsing_nba(self):
        self.assertEqual(
            self.parser.from_abbreviation(abbreviation="NBA"),
            League.NATIONAL_BASKETBALL_ASSOCIATION,
        )

    def test_parsing_aba(self):
        self.assertEqual(
            self.parser.from_abbreviation(abbreviation="ABA"),
            League.AMERICAN_BASKETBALL_ASSOCIATION,
        )

    def test_parsing_baa(self):
        self.assertEqual(
            self.parser.from_abbreviation(abbreviation="BAA"),
            League.BASKETBALL_ASSOCIATION_OF_AMERICA,
        )

    def test_from_abbreviations_when_abbreviations_is_none(self):
        self.assertEqual(
            self.parser.from_abbreviations(abbreviations=None),
            []
        )

    def test_from_abbreviations_parsing_single_league(self):
        self.assertEqual(
            self.parser.from_abbreviations(abbreviations="NBA"),
            [League.NATIONAL_BASKETBALL_ASSOCIATION],
        )

    def test_from_abbreviations_parsing_aba(self):
        self.assertEqual(
            self.parser.from_abbreviations(abbreviations="ABA"),
            [League.AMERICAN_BASKETBALL_ASSOCIATION],
        )

    def test_from_abbreviations_parsing_baa(self):
        self.assertEqual(
            self.parser.from_abbreviations(abbreviations="BAA"),
            [League.BASKETBALL_ASSOCIATION_OF_AMERICA],
        )

    def test_from_abbreviations_parsing_nba_aba_baa(self):
        self.assertEqual(
            self.parser.from_abbreviations(abbreviations="NBA/ABA/BAA"),
            [
                League.NATIONAL_BASKETBALL_ASSOCIATION,
                League.AMERICAN_BASKETBALL_ASSOCIATION,
                League.BASKETBALL_ASSOCIATION_OF_AMERICA,
            ],
        )



import json
import os
from unittest import TestCase

from basketball_reference_web_scraper import client
from basketball_reference_web_scraper.data import League, OutputType, OutputWriteOption


class TestSearchInMemory(TestCase):
    def test_search_ja(self):
        results = client.search(term="ja")
        self.assertGreaterEqual(498, len(results["players"]))

    def test_search_jaebaebae(self):
        results = client.search(term="jaebaebae")
        self.assertListEqual([], results["players"])

    def test_search_results_key(self):
        results = client.search(term="jaebaebae")
        self.assertListEqual(list(results), ["players"])

    def test_length_of_kobe_search_results(self):
        results = client.search(term="kobe")
        self.assertEqual(4, len(results["players"]))

    def test_players_in_kobe_search_results(self):
        results = client.search(term="kobe")
        self.assertListEqual(
            [
                {
                    "name": "Kobe Bryant",
                    "identifier": "bryanko01",
                    "leagues": {League.NATIONAL_BASKETBALL_ASSOCIATION}
                },
                {
                    "name": "Ruben Patterson",
                    "identifier": "patteru01",
                    "leagues": {League.NATIONAL_BASKETBALL_ASSOCIATION}
                },
                {
                    "name": "Dion Waiters",
                    "identifier": "waitedi01",
                    "leagues": {League.NATIONAL_BASKETBALL_ASSOCIATION}
                },
                {
                    "name": "Oleksandr Kobets",
                    "identifier": "kobetol01",
                    "leagues": set()
                }
            ],
            results["players"]
        )

    def test_exact_search_result(self):
        results = client.search(term="kobe bryant")
        self.assertEqual(
            [
                {
                    "name": "Kobe Bryant",
                    "identifier": "bryanko01",
                    "leagues": {League.NATIONAL_BASKETBALL_ASSOCIATION}
                }
            ],
            results["players"]
        )

    def test_large_search_pagination(self):
        results = client.search(term="a")
        self.assertGreaterEqual(len(results["players"]), 960)


class TestSearchJSONOutput(TestCase):
    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/ko_search.json",
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/ko_search.json",
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def test_ko_search_json_output_includes_expected_json_output(self):
        client.search(
            term="ko",
            output_type=OutputType.JSON,
            output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )
        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            output_data = json.load(output_file)
            expected_output_data = json.load(expected_output_file)
            for expected_data_row in expected_output_data:
                self.assertTrue(expected_data_row in output_data)


class TestSearchCSVOutput(TestCase):
    def setUp(self):
        self.output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/ko_search.csv",
        )
        self.expected_output_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/ko_search.csv",
        )

    def tearDown(self):
        os.remove(self.output_file_path)

    def test_ko_csv_output_search_includes_expected_csv_output(self):
        client.search(
            term="ko",
            output_type=OutputType.CSV,
            output_file_path=self.output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )

        with open(self.output_file_path, "r") as output_file, \
                open(self.expected_output_file_path, "r") as expected_output_file:
            output_data = output_file.readlines()
            expected_output_data = expected_output_file.readlines()
            for expected_data_row in expected_output_data:
                # TODO: @jaebradley this is freakin' gross but sets are not ordered (duh)
                # so serialization of the set of leagues will not be consistent.
                # Need to find a way to use an ordered set or something for this.
                # In the interim, ignore serialized sets of leagues - quick and dirty way
                # is to look for `-` (this ignores players with a `-` but I'll take the tradeoff for now)
                if "-" not in expected_data_row:
                    self.assertTrue(expected_data_row in output_data)



from unittest import TestCase

from basketball_reference_web_scraper import client
from basketball_reference_web_scraper.data import Position, Team


class TestPlayerSeasonTotals(TestCase):
    def test_2001_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2001)
        self.assertIsNotNone(player_season_totals)

    def test_2002_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2002)
        self.assertIsNotNone(player_season_totals)

    def test_2003_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2003)
        self.assertIsNotNone(player_season_totals)

    def test_2004_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2004)
        self.assertIsNotNone(player_season_totals)

    def test_2005_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2005)
        self.assertIsNotNone(player_season_totals)

    def test_2006_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2006)
        self.assertIsNotNone(player_season_totals)

    def test_2007_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2007)
        self.assertIsNotNone(player_season_totals)

    def test_2008_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2008)
        self.assertIsNotNone(player_season_totals)

    def test_2009_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2009)
        self.assertIsNotNone(player_season_totals)

    def test_2010_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2010)
        self.assertIsNotNone(player_season_totals)

    def test_2011_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2011)
        self.assertIsNotNone(player_season_totals)

    def test_2012_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2012)
        self.assertIsNotNone(player_season_totals)

    def test_2013_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2013)
        self.assertIsNotNone(player_season_totals)

    def test_2014_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2014)
        self.assertIsNotNone(player_season_totals)

    def test_2015_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2015)
        self.assertIsNotNone(player_season_totals)

    def test_2016_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2016)
        self.assertIsNotNone(player_season_totals)

    def test_2017_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2017)
        self.assertIsNotNone(player_season_totals)

    def test_2018_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2018)
        self.assertIsNotNone(player_season_totals)

    def test_2019_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2019)
        self.assertIsNotNone(player_season_totals)

    def test_avery_bradley_2019_player_season_totals(self):
        player_season_totals = client.players_season_totals(season_end_year=2019)
        clippers_avery_bradley = player_season_totals[66]

        self.assertEqual('bradlav01', clippers_avery_bradley["slug"])
        self.assertEqual("Avery Bradley", clippers_avery_bradley["name"])
        self.assertListEqual([Position.SHOOTING_GUARD], clippers_avery_bradley["positions"])
        self.assertEqual(28, clippers_avery_bradley["age"])
        self.assertEqual(Team.LOS_ANGELES_CLIPPERS, clippers_avery_bradley["team"])
        self.assertEqual(49, clippers_avery_bradley["games_played"])
        self.assertEqual(49, clippers_avery_bradley["games_started"])
        self.assertEqual(1463, clippers_avery_bradley["minutes_played"])
        self.assertEqual(161, clippers_avery_bradley["made_field_goals"])
        self.assertEqual(420, clippers_avery_bradley["attempted_field_goals"])
        self.assertEqual(58, clippers_avery_bradley["made_three_point_field_goals"])
        self.assertEqual(172, clippers_avery_bradley["attempted_three_point_field_goals"])
        self.assertEqual(20, clippers_avery_bradley["made_free_throws"])
        self.assertEqual(25, clippers_avery_bradley["attempted_free_throws"])
        self.assertEqual(35, clippers_avery_bradley["offensive_rebounds"])
        self.assertEqual(96, clippers_avery_bradley["defensive_rebounds"])
        self.assertEqual(96, clippers_avery_bradley["assists"])
        self.assertEqual(27, clippers_avery_bradley["steals"])
        self.assertEqual(16, clippers_avery_bradley["blocks"])
        self.assertEqual(61, clippers_avery_bradley["turnovers"])
        self.assertEqual(133, clippers_avery_bradley["personal_fouls"])
        self.assertEqual(400, clippers_avery_bradley["points"])



from datetime import datetime
from unittest import TestCase
import os

import basketball_reference_web_scraper.client as client
from basketball_reference_web_scraper.data import OutputWriteOption, OutputType, Team, PeriodType


class TestClient(TestCase):
    def test_schedules_from_2001(self):
        now = datetime.now()
        current_year = now.year

        for year in range(2001, current_year + 1):
            season_schedule = client.season_schedule(season_end_year=year)
            self.assertIsNotNone(season_schedule)

    def test_output_json_box_scores_to_file(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/foo.json",
        )
        client.player_box_scores(
            day=1,
            month=1,
            year=2001,
            output_type=OutputType.JSON,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE
        )

    def test_output_csv_box_scores_to_file(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2003_10_29_TOR_pbp.csv",
        )
        client.player_box_scores(
            day=1,
            month=1,
            year=2001,
            output_type=OutputType.CSV,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE
        )

    def test_output_json_box_scores_to_memory(self):
        january_first_box_scores = client.player_box_scores(
            day=1,
            month=1,
            year=2001,
            output_type=OutputType.JSON,
        )

        self.assertIsNotNone(january_first_box_scores)

    def test_2001_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2001)
        self.assertIsNotNone(schedule)

    def test_2002_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2002)
        self.assertIsNotNone(schedule)

    def test_2003_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2003)
        self.assertIsNotNone(schedule)

    def test_2004_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2004)
        self.assertIsNotNone(schedule)

    def test_2005_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2005)
        self.assertIsNotNone(schedule)

    def test_2006_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2006)
        self.assertIsNotNone(schedule)

    def test_2007_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2007)
        self.assertIsNotNone(schedule)

    def test_2008_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2008)
        self.assertIsNotNone(schedule)

    def test_2009_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2009)
        self.assertIsNotNone(schedule)

    def test_2010_season_schedule(self):
        player_season_totals = client.season_schedule(season_end_year=2010)
        self.assertIsNotNone(player_season_totals)

    def test_2011_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2011)
        self.assertIsNotNone(schedule)

    def test_2012_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2012)
        self.assertIsNotNone(schedule)

    def test_2013_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2013)
        self.assertIsNotNone(schedule)

    def test_2014_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2014)
        self.assertIsNotNone(schedule)

    def test_2015_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2015)
        self.assertIsNotNone(schedule)

    def test_2016_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2016)
        self.assertIsNotNone(schedule)

    def test_2017_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2017)
        self.assertIsNotNone(schedule)

    def test_2018_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2018)
        self.assertIsNotNone(schedule)

    def test_2019_season_schedule(self):
        schedule = client.season_schedule(season_end_year=2019)
        self.assertIsNotNone(schedule)

    def test_2019_player_advanced_season_totals(self):
        player_season_totals = client.players_advanced_season_totals(season_end_year=2019)
        self.assertIsNotNone(player_season_totals)
        self.assertTrue(len(player_season_totals) > 0)

    def test_BOS_2018_10_16_play_by_play(self):
        play_by_play = client.play_by_play(
            home_team=Team.BOSTON_CELTICS,
            day=16,
            month=10,
            year=2018,
        )
        self.assertIsNotNone(play_by_play)

    def test_BOS_2018_10_16_play_by_play_csv_to_file(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2018_10_16_BOS_pbp.csv",
        )
        client.play_by_play(
            home_team=Team.BOSTON_CELTICS,
            day=16,
            month=10,
            year=2018,
            output_type=OutputType.CSV,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )

    def test_overtime_play_by_play(self):
        play_by_play = client.play_by_play(
            home_team=Team.PORTLAND_TRAIL_BLAZERS,
            day=22,
            month=10,
            year=2018,
        )
        last_play = play_by_play[-1]
        self.assertIsNotNone(last_play)
        self.assertEqual(1, last_play["period"])
        self.assertEqual(PeriodType.OVERTIME, last_play["period_type"])

    def test_overtime_play_by_play_to_csv(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2018_10_22_POR_pbp.csv",
        )
        client.play_by_play(
            home_team=Team.PORTLAND_TRAIL_BLAZERS,
            day=22,
            month=10,
            year=2018,
            output_type=OutputType.CSV,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )

    def test_BOS_2018_10_16_play_by_play_json_to_file(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2018_10_16_BOS_pbp.json",
        )
        client.play_by_play(
            home_team=Team.BOSTON_CELTICS,
            day=16,
            month=10,
            year=2018,
            output_type=OutputType.JSON,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )

    def test_overtime_play_by_play_to_json_file(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2018_10_22_POR_pbp.json",
        )
        client.play_by_play(
            home_team=Team.PORTLAND_TRAIL_BLAZERS,
            day=22,
            month=10,
            year=2018,
            output_type=OutputType.JSON,
            output_file_path=output_file_path,
            output_write_option=OutputWriteOption.WRITE,
        )



from unittest import TestCase

from basketball_reference_web_scraper.http_client import SEARCH_RESULT_RESOURCE_LOCATION_REGEX
from basketball_reference_web_scraper.parsers import ResourceLocationParser


class TestResourceLocationParser(TestCase):
    def setUp(self):
        self.parser = ResourceLocationParser(resource_location_regex=SEARCH_RESULT_RESOURCE_LOCATION_REGEX)

    def test_parse_players_resource_type(self):
        self.assertEqual(
            self.parser.parse_resource_type(
                resource_location="https://www.basketball-reference.com/players/k/koperbu01.html"
            ),
            "players"
        )

    def test_parse_coaches_resource_type(self):
        self.assertEqual(
            self.parser.parse_resource_type(
                resource_location="https://www.basketball-reference.com/coaches/vanbrbu01c.html"
            ),
            "coaches"
        )

    def test_parse_executives_resource_type(self):
        self.assertEqual(
            self.parser.parse_resource_type(
                resource_location="https://www.basketball-reference.com/executives/vanbrbu01x.html"
            ),
            "executives",
        )

    def test_parse_players_resource_identifier(self):
        self.assertEqual(
            self.parser.parse_resource_identifier(
                resource_location="https://www.basketball-reference.com/players/k/koperbu01.html"
            ),
            "koperbu01"
        )

    def test_parse_coaches_resource_identifier(self):
        self.assertEqual(
            self.parser.parse_resource_identifier(
                resource_location="https://www.basketball-reference.com/coaches/vanbrbu01c.html"
            ),
            "vanbrbu01c"
        )

    def test_parse_executives_resource_identifier(self):
        self.assertEqual(
            self.parser.parse_resource_identifier(
                resource_location="https://www.basketball-reference.com/executives/vanbrbu01x.html"
            ),
            "vanbrbu01x",
        )



from unittest import TestCase
from unittest.mock import MagicMock

from basketball_reference_web_scraper.parsers import TeamTotalsParser, TeamAbbreviationParser
from basketball_reference_web_scraper.data import Outcome, TeamTotal, TEAM_ABBREVIATIONS_TO_TEAM


class TestTeamTotalsParser(TestCase):
    def setUp(self):
        self.parser = TeamTotalsParser(
            team_abbreviation_parser=TeamAbbreviationParser(
                abbreviations_to_teams=TEAM_ABBREVIATIONS_TO_TEAM
            )
        )

    def test_parse_none_outcome_when_points_are_same(self):
        team_totals = TeamTotal(team_abbreviation="BOS", totals=MagicMock(points="100"))
        opposing_team_totals = TeamTotal(team_abbreviation="GSW", totals=MagicMock(points="100"))
        self.assertIsNone(
            self.parser.parse_totals(team_totals=team_totals, opposing_team_totals=opposing_team_totals)["outcome"]
        )



import re
from datetime import datetime

import pytz

from basketball_reference_web_scraper.data import PeriodType, Outcome
from basketball_reference_web_scraper.utilities import str_to_int, str_to_float

PLAYER_SEASON_BOX_SCORES_GAME_DATE_FORMAT = '%Y-%m-%d'
PLAYER_SEASON_BOX_SCORES_OUTCOME_REGEX = '(?P<outcome_abbreviation>W|L) \((?P<margin_of_victory>[^)]+)\)'
SEARCH_RESULT_NAME_REGEX = '(?P<name>^[^\(]+)'


class TeamAbbreviationParser:
    def __init__(self, abbreviations_to_teams):
        self.abbreviations_to_teams = abbreviations_to_teams

    def from_abbreviation(self, abbreviation):
        return self.abbreviations_to_teams.get(abbreviation)


class PositionAbbreviationParser:
    def __init__(self, abbreviations_to_positions):
        self.abbreviations_to_positions = abbreviations_to_positions

    def from_abbreviation(self, abbreviation):
        return self.abbreviations_to_positions.get(abbreviation)

    def from_abbreviations(self, abbreviations):
        parsed_positions = list(
            map(
                lambda position_abbreviation: self.from_abbreviation(position_abbreviation),
                abbreviations.split("-")
            )
        )
        return [position for position in parsed_positions if position is not None]


class LocationAbbreviationParser:
    def __init__(self, abbreviations_to_locations):
        self.abbreviations_to_locations = abbreviations_to_locations

    def from_abbreviation(self, abbreviation):
        location = self.abbreviations_to_locations.get(abbreviation)
        if location is None:
            raise ValueError("Unknown symbol: {location}".format(location=location))

        return location


class OutcomeAbbreviationParser:
    def __init__(self, abbreviations_to_outcomes):
        self.abbreviations_to_outcomes = abbreviations_to_outcomes

    def from_abbreviation(self, abbreviation):
        outcome = self.abbreviations_to_outcomes.get(abbreviation)
        if outcome is None:
            raise ValueError("Unknown symbol: {outcome}".format(outcome=outcome))

        return outcome


class LeagueAbbreviationParser:
    def __init__(self, abbreviations_to_league):
        self.abbreviations_to_league = abbreviations_to_league

    def from_abbreviation(self, abbreviation):
        league = self.abbreviations_to_league.get(abbreviation)
        if league is None:
            raise ValueError("Unknown league abbreviation: {abbreviation}".format(abbreviation=abbreviation))

        return league

    def from_abbreviations(self, abbreviations):
        if abbreviations is None:
            return []

        return [
            self.from_abbreviation(abbreviation=league_abbreviation)
            for league_abbreviation in abbreviations.split("/")
        ]


class PlayerBoxScoreOutcomeParser:
    def __init__(self,
                 outcome_abbreviation_parser,
                 formatted_outcome_regex=PLAYER_SEASON_BOX_SCORES_OUTCOME_REGEX,
                 outcome_abbreviation_regex_group_name='outcome_abbreviation',
                 margin_of_victory_regex_group_name='margin_of_victory',
                 ):
        self.outcome_abbreviation_parser = outcome_abbreviation_parser
        self.formatted_outcome_regex = formatted_outcome_regex
        self.outcome_abbreviation_regex_group_name = outcome_abbreviation_regex_group_name
        self.margin_of_victory_regex_group_name = margin_of_victory_regex_group_name

    def search_formatted_outcome(self, formatted_outcome):
        return re.search(self.formatted_outcome_regex, formatted_outcome)

    def parse_outcome_abbreviation(self, formatted_outcome):
        return self.search_formatted_outcome(formatted_outcome=formatted_outcome) \
            .group(self.outcome_abbreviation_regex_group_name)

    def parse_outcome(self, formatted_outcome):
        return self.outcome_abbreviation_parser.from_abbreviation(
            abbreviation=self.parse_outcome_abbreviation(formatted_outcome=formatted_outcome)
        )

    def parse_margin_of_victory(self, formatted_outcome):
        return int(
            self.search_formatted_outcome(formatted_outcome=formatted_outcome)
                .group(self.margin_of_victory_regex_group_name)
        )


class SecondsPlayedParser:
    def parse(self, formatted_playing_time):
        if formatted_playing_time == "":
            return 0

        # It seems like basketball reference formats everything in MM:SS
        # even when the playing time is greater than 59 minutes, 59 seconds.
        #
        # Because of this, we can't use strptime / %M as valid values are 0-59.
        # So have to parse time by splitting on ":" and assuming that
        # the first part is the minute part and the second part is the seconds part
        time_parts = formatted_playing_time.split(":")
        minutes_played = time_parts[0]
        seconds_played = time_parts[1]
        return 60 * int(minutes_played) + int(seconds_played)


class PeriodDetailsParser:
    def __init__(self, regulation_periods_count):
        self.regulation_periods_count = regulation_periods_count

    def is_overtime(self, period_count):
        return period_count > self.regulation_periods_count

    def parse_period_number(self, period_count):
        if self.is_overtime(period_count=period_count):
            return period_count - self.regulation_periods_count

        return period_count

    def parse_period_type(self, period_count):
        if self.is_overtime(period_count=period_count):
            return PeriodType.OVERTIME

        return PeriodType.QUARTER


class PeriodTimestampParser:
    def __init__(self, timestamp_format):
        self.timestamp_format = timestamp_format

    def to_seconds(self, timestamp):
        dt = datetime.strptime(timestamp, self.timestamp_format)
        return float(
            (dt.minute * 60) + dt.second + (dt.microsecond / 1000000)
        )


class ScoresParser:
    def __init__(
            self,
            scores_regex,
            away_team_score_group_name='away_team_score',
            home_team_score_group_name='home_team_score',
    ):
        self.scores_regex = scores_regex
        self.away_team_score_group_name = away_team_score_group_name
        self.home_team_score_group_name = home_team_score_group_name

    def parse_scores(self, formatted_scores):
        return re.search(self.scores_regex, formatted_scores)

    def parse_away_team_score(self, formatted_scores):
        return int(
            self.parse_scores(formatted_scores=formatted_scores)
                .group(self.away_team_score_group_name)
        )

    def parse_home_team_score(self, formatted_scores):
        return int(
            self.parse_scores(formatted_scores=formatted_scores)
                .group(self.home_team_score_group_name)
        )


class TeamNameParser:
    def __init__(self, team_names_to_teams):
        self.team_names_to_teams = team_names_to_teams

    def parse_team_name(self, team_name):
        return self.team_names_to_teams[team_name.strip().upper()]


class ScheduledStartTimeParser:
    def __init__(self, time_zone=pytz.utc):
        self.time_zone = time_zone

    def parse_start_time(self, formatted_date, formatted_time_of_day):
        if formatted_time_of_day is not None and formatted_time_of_day not in ["", " "]:
            # Starting in 2018, the start times had a "p" or "a" appended to the end
            # Between 2001 and 2017, the start times had a "pm" or "am"
            #
            # https://www.basketball-reference.com/leagues/NBA_2018_games.html
            # vs.
            # https://www.basketball-reference.com/leagues/NBA_2001_games.html
            is_prior_format = formatted_time_of_day[-2:] == "am" or formatted_time_of_day[-2:] == "pm"

            # If format contains only "p" or "a" add an "m" so it can be parsed by datetime module
            if is_prior_format:
                combined_formatted_time = formatted_date + " " + formatted_time_of_day
            else:
                combined_formatted_time = formatted_date + " " + formatted_time_of_day + "m"

            if is_prior_format:
                start_time = datetime.strptime(combined_formatted_time, "%a, %b %d, %Y %I:%M %p")
            else:
                start_time = datetime.strptime(combined_formatted_time, "%a, %b %d, %Y %I:%M%p")
        else:
            start_time = datetime.strptime(formatted_date, "%a, %b %d, %Y")

        # All basketball reference times seem to be in Eastern
        est = pytz.timezone("US/Eastern")
        localized_start_time = est.localize(start_time)
        return localized_start_time.astimezone(self.time_zone)


class SearchResultNameParser:
    def __init__(self, search_result_name_regex=SEARCH_RESULT_NAME_REGEX, result_name_regex_group_name="name"):
        self.search_result_name_regex = search_result_name_regex
        self.result_name_regex_group_name = result_name_regex_group_name

    def parse(self, search_result_name):
        return re.search(self.search_result_name_regex, search_result_name) \
            .group(self.result_name_regex_group_name) \
            .strip()


class ResourceLocationParser:
    def __init__(self,
                 resource_location_regex,
                 resource_type_regex_group_name="resource_type",
                 resource_identifier_regex_group_name="resource_identifier"):
        self.resource_location_regex = resource_location_regex
        self.resource_type_regex_group_name = resource_type_regex_group_name
        self.resource_identifier_regex_group_name = resource_identifier_regex_group_name

    def search(self, resource_location):
        return re.search(self.resource_location_regex, resource_location)

    def parse_resource_type(self, resource_location):
        return self.search(resource_location=resource_location).group(self.resource_type_regex_group_name)

    def parse_resource_identifier(self, resource_location):
        return self.search(resource_location=resource_location).group(self.resource_identifier_regex_group_name)


class ScheduledGamesParser:
    def __init__(self, start_time_parser, team_name_parser):
        self.start_time_parser = start_time_parser
        self.team_name_parser = team_name_parser

    def parse_games(self, games):
        return [
            {
                "start_time": self.start_time_parser.parse_start_time(
                    formatted_date=game.start_date,
                    formatted_time_of_day=game.start_time_of_day,
                ),
                "away_team": self.team_name_parser.parse_team_name(team_name=game.away_team_name),
                "home_team": self.team_name_parser.parse_team_name(team_name=game.home_team_name),
                "away_team_score": str_to_int(value=game.away_team_score, default=None),
                "home_team_score": str_to_int(value=game.home_team_score, default=None),
            }
            for game in games
        ]


class PlayerAdvancedSeasonTotalsParser:
    def __init__(self, position_abbreviation_parser, team_abbreviation_parser):
        self.position_abbreviation_parser = position_abbreviation_parser
        self.team_abbreviation_parser = team_abbreviation_parser

    def parse(self, totals):
        return [
            {
                "slug": str(total.slug),
                "name": str(total.name),
                "positions": self.position_abbreviation_parser.from_abbreviations(total.position_abbreviations),
                "age": str_to_int(total.age, default=None),
                "team": self.team_abbreviation_parser.from_abbreviation(total.team_abbreviation),
                "games_played": str_to_int(total.games_played),
                "minutes_played": str_to_int(total.minutes_played),
                "player_efficiency_rating": str_to_float(total.player_efficiency_rating),
                "true_shooting_percentage": str_to_float(total.true_shooting_percentage),
                "three_point_attempt_rate": str_to_float(total.three_point_attempt_rate),
                "free_throw_attempt_rate": str_to_float(total.free_throw_attempt_rate),
                "offensive_rebound_percentage": str_to_float(total.offensive_rebound_percentage),
                "defensive_rebound_percentage": str_to_float(total.defensive_rebound_percentage),
                "total_rebound_percentage": str_to_float(total.total_rebound_percentage),
                "assist_percentage": str_to_float(total.assist_percentage),
                "steal_percentage": str_to_float(total.steal_percentage),
                "block_percentage": str_to_float(total.block_percentage),
                "turnover_percentage": str_to_float(total.turnover_percentage),
                "usage_percentage": str_to_float(total.usage_percentage),
                "offensive_win_shares": str_to_float(total.offensive_win_shares),
                "defensive_win_shares": str_to_float(total.defensive_win_shares),
                "win_shares": str_to_float(total.win_shares),
                "win_shares_per_48_minutes": str_to_float(total.win_shares_per_48_minutes),
                "offensive_box_plus_minus": str_to_float(total.offensive_plus_minus),
                "defensive_box_plus_minus": str_to_float(total.defensive_plus_minus),
                "box_plus_minus": str_to_float(total.plus_minus),
                "value_over_replacement_player": str_to_float(total.value_over_replacement_player),
                "is_combined_totals": total.is_combined_totals,
            } for total in totals
        ]


class PlayerSeasonTotalsParser:
    def __init__(self, position_abbreviation_parser, team_abbreviation_parser):
        self.position_abbreviation_parser = position_abbreviation_parser
        self.team_abbreviation_parser = team_abbreviation_parser

    def parse(self, totals):
        return [
            {
                "slug": str(total.slug),
                "name": str(total.name),
                "positions": self.position_abbreviation_parser.from_abbreviations(total.position_abbreviations),
                "age": str_to_int(total.age, default=None),
                "team": self.team_abbreviation_parser.from_abbreviation(total.team_abbreviation),
                "games_played": str_to_int(total.games_played),
                "games_started": str_to_int(total.games_started),
                "minutes_played": str_to_int(total.minutes_played),
                "made_field_goals": str_to_int(total.made_field_goals),
                "attempted_field_goals": str_to_int(total.attempted_field_goals),
                "made_three_point_field_goals": str_to_int(total.made_three_point_field_goals),
                "attempted_three_point_field_goals": str_to_int(total.attempted_three_point_field_goals),
                "made_free_throws": str_to_int(total.made_free_throws),
                "attempted_free_throws": str_to_int(total.attempted_free_throws),
                "offensive_rebounds": str_to_int(total.offensive_rebounds),
                "defensive_rebounds": str_to_int(total.defensive_rebounds),
                "assists": str_to_int(total.assists),
                "steals": str_to_int(total.steals),
                "blocks": str_to_int(total.blocks),
                "turnovers": str_to_int(total.turnovers),
                "personal_fouls": str_to_int(total.personal_fouls),
                "points": str_to_int(total.points),
            } for total in totals
        ]


class TeamTotalsParser:
    def __init__(self, team_abbreviation_parser):
        self.team_abbreviation_parser = team_abbreviation_parser

    def parse(self, first_team_totals, second_team_totals):
        return [
            self.parse_totals(
                team_totals=first_team_totals,
                opposing_team_totals=second_team_totals,
            ),
            self.parse_totals(
                team_totals=second_team_totals,
                opposing_team_totals=first_team_totals,
            ),
        ]

    def parse_totals(self, team_totals, opposing_team_totals):
        current_team = self.team_abbreviation_parser.from_abbreviation(team_totals.team_abbreviation)

        if str_to_int(team_totals.points) > str_to_int(opposing_team_totals.points):
            outcome = Outcome.WIN
        elif str_to_int(team_totals.points) < str_to_int(opposing_team_totals.points):
            outcome = Outcome.LOSS
        else:
            outcome = None

        return {
            "team": current_team,
            "outcome": outcome,
            "minutes_played": str_to_int(team_totals.minutes_played),
            "made_field_goals": str_to_int(team_totals.made_field_goals),
            "attempted_field_goals": str_to_int(team_totals.attempted_field_goals),
            "made_three_point_field_goals": str_to_int(team_totals.made_three_point_field_goals),
            "attempted_three_point_field_goals": str_to_int(team_totals.attempted_three_point_field_goals),
            "made_free_throws": str_to_int(team_totals.made_free_throws),
            "attempted_free_throws": str_to_int(team_totals.attempted_free_throws),
            "offensive_rebounds": str_to_int(team_totals.offensive_rebounds),
            "defensive_rebounds": str_to_int(team_totals.defensive_rebounds),
            "assists": str_to_int(team_totals.assists),
            "steals": str_to_int(team_totals.steals),
            "blocks": str_to_int(team_totals.blocks),
            "turnovers": str_to_int(team_totals.turnovers),
            "personal_fouls": str_to_int(team_totals.personal_fouls),
            "points": str_to_int(team_totals.points),
        }


class PlayerBoxScoresParser:
    def __init__(self, team_abbreviation_parser, location_abbreviation_parser, outcome_abbreviation_parser,
                 seconds_played_parser):
        self.team_abbreviation_parser = team_abbreviation_parser
        self.location_abbreviation_parser = location_abbreviation_parser
        self.outcome_abbreviation_parser = outcome_abbreviation_parser
        self.seconds_played_parser = seconds_played_parser

    def parse(self, box_scores):
        return [
            {
                "slug": str(box_score.slug),
                "name": str(box_score.name),
                "team": self.team_abbreviation_parser.from_abbreviation(box_score.team_abbreviation),
                "location": self.location_abbreviation_parser.from_abbreviation(box_score.location_abbreviation),
                "opponent": self.team_abbreviation_parser.from_abbreviation(box_score.opponent_abbreviation),
                "outcome": self.outcome_abbreviation_parser.from_abbreviation(box_score.outcome),
                "seconds_played": self.seconds_played_parser.parse(box_score.playing_time),
                "made_field_goals": str_to_int(box_score.made_field_goals),
                "attempted_field_goals": str_to_int(box_score.attempted_field_goals),
                "made_three_point_field_goals": str_to_int(box_score.made_three_point_field_goals),
                "attempted_three_point_field_goals": str_to_int(box_score.attempted_three_point_field_goals),
                "made_free_throws": str_to_int(box_score.made_free_throws),
                "attempted_free_throws": str_to_int(box_score.attempted_free_throws),
                "offensive_rebounds": str_to_int(box_score.offensive_rebounds),
                "defensive_rebounds": str_to_int(box_score.defensive_rebounds),
                "assists": str_to_int(box_score.assists),
                "steals": str_to_int(box_score.steals),
                "blocks": str_to_int(box_score.blocks),
                "turnovers": str_to_int(box_score.turnovers),
                "personal_fouls": str_to_int(box_score.personal_fouls),
                "game_score": str_to_float(box_score.game_score),
            } for box_score in box_scores
        ]


class PlayerSeasonBoxScoresParser:
    def __init__(self, team_abbreviation_parser, location_abbreviation_parser, outcome_parser, seconds_played_parser):
        self.team_abbreviation_parser = team_abbreviation_parser
        self.location_abbreviation_parser = location_abbreviation_parser
        self.outcome_parser = outcome_parser
        self.seconds_played_parser = seconds_played_parser

    def parse(self, box_scores):
        return [
            {
                "date": datetime.strptime(str(box_score.date), "%Y-%m-%d").date(),
                "team": self.team_abbreviation_parser.from_abbreviation(box_score.team_abbreviation),
                "location": self.location_abbreviation_parser.from_abbreviation(box_score.location_abbreviation),
                "opponent": self.team_abbreviation_parser.from_abbreviation(box_score.opponent_abbreviation),
                "outcome": self.outcome_parser.parse_outcome(formatted_outcome=box_score.outcome),
                "seconds_played": self.seconds_played_parser.parse(box_score.playing_time),
                "made_field_goals": str_to_int(box_score.made_field_goals),
                "attempted_field_goals": str_to_int(box_score.attempted_field_goals),
                "made_three_point_field_goals": str_to_int(box_score.made_three_point_field_goals),
                "attempted_three_point_field_goals": str_to_int(box_score.attempted_three_point_field_goals),
                "made_free_throws": str_to_int(box_score.made_free_throws),
                "attempted_free_throws": str_to_int(box_score.attempted_free_throws),
                "offensive_rebounds": str_to_int(box_score.offensive_rebounds),
                "defensive_rebounds": str_to_int(box_score.defensive_rebounds),
                "assists": str_to_int(box_score.assists),
                "steals": str_to_int(box_score.steals),
                "blocks": str_to_int(box_score.blocks),
                "turnovers": str_to_int(box_score.turnovers),
                "personal_fouls": str_to_int(box_score.personal_fouls),
                "points_scored": str_to_int(box_score.points_scored),
                "game_score": str_to_float(box_score.game_score),
                "plus_minus": str_to_int(box_score.plus_minus),
            } for box_score in box_scores
            if box_score.is_active
        ]


class PlayByPlaysParser:
    def __init__(self, period_details_parser, period_timestamp_parser, scores_parser):
        self.period_details_parser = period_details_parser
        self.period_timestamp_parser = period_timestamp_parser
        self.scores_parser = scores_parser

    def parse(self, play_by_plays, away_team, home_team):
        current_period = 0
        result = []
        for play_by_play in play_by_plays:
            if play_by_play.is_start_of_period:
                current_period += 1
            elif play_by_play.has_play_by_play_data:
                result.append(self.format_data(
                    current_period=current_period,
                    play_by_play=play_by_play,
                    away_team=away_team,
                    home_team=home_team,
                ))
        return result

    def format_data(self, current_period, play_by_play, away_team, home_team):
        return {
            "period": self.period_details_parser.parse_period_number(period_count=current_period),
            "period_type": self.period_details_parser.parse_period_type(period_count=current_period),
            "remaining_seconds_in_period": self.period_timestamp_parser.to_seconds(timestamp=play_by_play.timestamp),
            "relevant_team": away_team if play_by_play.is_away_team_play else home_team,
            "away_team": away_team,
            "home_team": home_team,
            "away_score": self.scores_parser.parse_away_team_score(formatted_scores=play_by_play.formatted_scores),
            "home_score": self.scores_parser.parse_home_team_score(formatted_scores=play_by_play.formatted_scores),
            "description": play_by_play.away_team_play_description
            if play_by_play.is_away_team_play
            else play_by_play.home_team_play_description,
        }


class SearchResultsParser:
    def __init__(self, search_result_name_parser, search_result_location_parser, league_abbreviation_parser):
        self.search_result_name_parser = search_result_name_parser
        self.search_result_location_parser = search_result_location_parser
        self.league_abbreviation_parser = league_abbreviation_parser

    def parse(self, nba_aba_baa_players):
        return {
            "players": [
                {
                    "name": self.search_result_name_parser.parse(search_result_name=result.resource_name),
                    "identifier": self.search_result_location_parser.parse_resource_identifier(
                        resource_location=result.resource_location
                    ),
                    "leagues": set(
                        self.league_abbreviation_parser.from_abbreviations(
                            abbreviations=result.league_abbreviations
                        )
                    ),
                } for result in nba_aba_baa_players
            ]
        }


class PlayerDataParser:
    def __init__(self, search_result_location_parser, league_abbreviation_parser):
        self.search_result_location_parser = search_result_location_parser
        self.league_abbreviation_parser = league_abbreviation_parser

    def parse(self, player):
        return {
            "name": player.name,
            "identifier": self.search_result_location_parser.parse_resource_identifier(
                resource_location=player.resource_location
            ),
            "leagues": set(
                (
                    self.league_abbreviation_parser.from_abbreviation(abbreviation=abbreviation)
                    for abbreviation in player.league_abbreviations
                )
            )
        }



from enum import Enum


class Location(Enum):
    HOME = "HOME"
    AWAY = "AWAY"


class Outcome(Enum):
    WIN = "WIN"
    LOSS = "LOSS"


class Team(Enum):
    ATLANTA_HAWKS = "ATLANTA HAWKS"
    BOSTON_CELTICS = "BOSTON CELTICS"
    BROOKLYN_NETS = "BROOKLYN NETS"
    CHARLOTTE_HORNETS = "CHARLOTTE HORNETS"
    CHICAGO_BULLS = "CHICAGO BULLS"
    CLEVELAND_CAVALIERS = "CLEVELAND CAVALIERS"
    DALLAS_MAVERICKS = "DALLAS MAVERICKS"
    DENVER_NUGGETS = "DENVER NUGGETS"
    DETROIT_PISTONS = "DETROIT PISTONS"
    GOLDEN_STATE_WARRIORS = "GOLDEN STATE WARRIORS"
    HOUSTON_ROCKETS = "HOUSTON ROCKETS"
    INDIANA_PACERS = "INDIANA PACERS"
    LOS_ANGELES_CLIPPERS = "LOS ANGELES CLIPPERS"
    LOS_ANGELES_LAKERS = "LOS ANGELES LAKERS"
    MEMPHIS_GRIZZLIES = "MEMPHIS GRIZZLIES"
    MIAMI_HEAT = "MIAMI HEAT"
    MILWAUKEE_BUCKS = "MILWAUKEE BUCKS"
    MINNESOTA_TIMBERWOLVES = "MINNESOTA TIMBERWOLVES"
    NEW_ORLEANS_PELICANS = "NEW ORLEANS PELICANS"
    NEW_YORK_KNICKS = "NEW YORK KNICKS"
    OKLAHOMA_CITY_THUNDER = "OKLAHOMA CITY THUNDER"
    ORLANDO_MAGIC = "ORLANDO MAGIC"
    PHILADELPHIA_76ERS = "PHILADELPHIA 76ERS"
    PHOENIX_SUNS = "PHOENIX SUNS"
    PORTLAND_TRAIL_BLAZERS = "PORTLAND TRAIL BLAZERS"
    SACRAMENTO_KINGS = "SACRAMENTO KINGS"
    SAN_ANTONIO_SPURS = "SAN ANTONIO SPURS"
    TORONTO_RAPTORS = "TORONTO RAPTORS"
    UTAH_JAZZ = "UTAH JAZZ"
    WASHINGTON_WIZARDS = "WASHINGTON WIZARDS"

    # DEPRECATED TEAMS
    CHARLOTTE_BOBCATS = "CHARLOTTE BOBCATS"
    NEW_JERSEY_NETS = "NEW JERSEY NETS"
    NEW_ORLEANS_HORNETS = "NEW ORLEANS HORNETS"
    NEW_ORLEANS_OKLAHOMA_CITY_HORNETS = "NEW ORLEANS/OKLAHOMA CITY HORNETS"
    SEATTLE_SUPERSONICS = "SEATTLE SUPERSONICS"
    VANCOUVER_GRIZZLIES = "VANCOUVER GRIZZLIES"


class OutputType(Enum):
    JSON = "JSON"
    CSV = "CSV"


class OutputWriteOption(Enum):
    WRITE = "w"
    CREATE_AND_WRITE = "w+"
    APPEND = "a"
    APPEND_AND_WRITE = "a+"


class Position(Enum):
    POINT_GUARD = "POINT GUARD"
    SHOOTING_GUARD = "SHOOTING GUARD"
    SMALL_FORWARD = "SMALL FORWARD"
    POWER_FORWARD = "POWER FORWARD"
    CENTER = "CENTER"
    FORWARD = "FORWARD"
    GUARD = "GUARD"


class PeriodType(Enum):
    QUARTER = "QUARTER"
    OVERTIME = "OVERTIME"


class League(Enum):
    NATIONAL_BASKETBALL_ASSOCIATION = "NATIONAL_BASKETBALL_ASSOCIATION"
    AMERICAN_BASKETBALL_ASSOCIATION = "AMERICAN_BASKETBALL_ASSOCIATION"
    BASKETBALL_ASSOCIATION_OF_AMERICA = "BASKETBALL_ASSOCIATION_OF_AMERICA"


TEAM_ABBREVIATIONS_TO_TEAM = {
    'ATL': Team.ATLANTA_HAWKS,
    'BOS': Team.BOSTON_CELTICS,
    'BRK': Team.BROOKLYN_NETS,
    'CHI': Team.CHICAGO_BULLS,
    'CHO': Team.CHARLOTTE_HORNETS,
    'CLE': Team.CLEVELAND_CAVALIERS,
    'DAL': Team.DALLAS_MAVERICKS,
    'DEN': Team.DENVER_NUGGETS,
    'DET': Team.DETROIT_PISTONS,
    'GSW': Team.GOLDEN_STATE_WARRIORS,
    'HOU': Team.HOUSTON_ROCKETS,
    'IND': Team.INDIANA_PACERS,
    'LAC': Team.LOS_ANGELES_CLIPPERS,
    'LAL': Team.LOS_ANGELES_LAKERS,
    'MEM': Team.MEMPHIS_GRIZZLIES,
    'MIA': Team.MIAMI_HEAT,
    'MIL': Team.MILWAUKEE_BUCKS,
    'MIN': Team.MINNESOTA_TIMBERWOLVES,
    'NOP': Team.NEW_ORLEANS_PELICANS,
    'NYK': Team.NEW_YORK_KNICKS,
    'OKC': Team.OKLAHOMA_CITY_THUNDER,
    'ORL': Team.ORLANDO_MAGIC,
    'PHI': Team.PHILADELPHIA_76ERS,
    'PHO': Team.PHOENIX_SUNS,
    'POR': Team.PORTLAND_TRAIL_BLAZERS,
    'SAC': Team.SACRAMENTO_KINGS,
    'SAS': Team.SAN_ANTONIO_SPURS,
    'TOR': Team.TORONTO_RAPTORS,
    'UTA': Team.UTAH_JAZZ,
    'WAS': Team.WASHINGTON_WIZARDS,

    # DEPRECATED TEAMS
    'NJN': Team.NEW_JERSEY_NETS,
    'NOH': Team.NEW_ORLEANS_HORNETS,
    'NOK': Team.NEW_ORLEANS_OKLAHOMA_CITY_HORNETS,
    'CHA': Team.CHARLOTTE_BOBCATS,
    'CHH': Team.CHARLOTTE_HORNETS,
    'SEA': Team.SEATTLE_SUPERSONICS,
    'VAN': Team.VANCOUVER_GRIZZLIES,
}

TEAM_TO_TEAM_ABBREVIATION = {v: k for k, v in TEAM_ABBREVIATIONS_TO_TEAM.items()}
TEAM_TO_TEAM_ABBREVIATION[Team.CHARLOTTE_HORNETS] = "CHO"

TEAM_NAME_TO_TEAM = {
    "ATLANTA HAWKS": Team.ATLANTA_HAWKS,
    "BOSTON CELTICS": Team.BOSTON_CELTICS,
    "BROOKLYN NETS": Team.BROOKLYN_NETS,
    "CHARLOTTE HORNETS": Team.CHARLOTTE_HORNETS,
    "CHICAGO BULLS": Team.CHICAGO_BULLS,
    "CLEVELAND CAVALIERS": Team.CLEVELAND_CAVALIERS,
    "DALLAS MAVERICKS": Team.DALLAS_MAVERICKS,
    "DENVER NUGGETS": Team.DENVER_NUGGETS,
    "DETROIT PISTONS": Team.DETROIT_PISTONS,
    "GOLDEN STATE WARRIORS": Team.GOLDEN_STATE_WARRIORS,
    "HOUSTON ROCKETS": Team.HOUSTON_ROCKETS,
    "INDIANA PACERS": Team.INDIANA_PACERS,
    "LOS ANGELES CLIPPERS": Team.LOS_ANGELES_CLIPPERS,
    "LOS ANGELES LAKERS": Team.LOS_ANGELES_LAKERS,
    "MEMPHIS GRIZZLIES": Team.MEMPHIS_GRIZZLIES,
    "MIAMI HEAT": Team.MIAMI_HEAT,
    "MILWAUKEE BUCKS": Team.MILWAUKEE_BUCKS,
    "MINNESOTA TIMBERWOLVES": Team.MINNESOTA_TIMBERWOLVES,
    "NEW ORLEANS PELICANS": Team.NEW_ORLEANS_PELICANS,
    "NEW YORK KNICKS": Team.NEW_YORK_KNICKS,
    "OKLAHOMA CITY THUNDER": Team.OKLAHOMA_CITY_THUNDER,
    "ORLANDO MAGIC": Team.ORLANDO_MAGIC,
    "PHILADELPHIA 76ERS": Team.PHILADELPHIA_76ERS,
    "PHOENIX SUNS": Team.PHOENIX_SUNS,
    "PORTLAND TRAIL BLAZERS": Team.PORTLAND_TRAIL_BLAZERS,
    "SACRAMENTO KINGS": Team.SACRAMENTO_KINGS,
    "SAN ANTONIO SPURS": Team.SAN_ANTONIO_SPURS,
    "TORONTO RAPTORS": Team.TORONTO_RAPTORS,
    "UTAH JAZZ": Team.UTAH_JAZZ,
    "WASHINGTON WIZARDS": Team.WASHINGTON_WIZARDS,

    # DEPRECATED TEAMS
    "CHARLOTTE BOBCATS": Team.CHARLOTTE_BOBCATS,
    "NEW JERSEY NETS": Team.NEW_JERSEY_NETS,
    "NEW ORLEANS HORNETS": Team.NEW_ORLEANS_HORNETS,
    "NEW ORLEANS/OKLAHOMA CITY HORNETS": Team.NEW_ORLEANS_OKLAHOMA_CITY_HORNETS,
    "SEATTLE SUPERSONICS": Team.SEATTLE_SUPERSONICS,
    "VANCOUVER GRIZZLIES": Team.VANCOUVER_GRIZZLIES,
}

POSITION_ABBREVIATIONS_TO_POSITION = {
    "PG": Position.POINT_GUARD,
    "SG": Position.SHOOTING_GUARD,
    "SF": Position.SMALL_FORWARD,
    "PF": Position.POWER_FORWARD,
    "C": Position.CENTER,
    "F": Position.FORWARD,
    "G": Position.GUARD,
}


LOCATION_ABBREVIATIONS_TO_POSITION = {
    "": Location.HOME,
    "@": Location.AWAY,
}


OUTCOME_ABBREVIATIONS_TO_OUTCOME = {
    "W": Outcome.WIN,
    "L": Outcome.LOSS,
}

LEAGUE_ABBREVIATIONS_TO_LEAGUE = {
    "NBA": League.NATIONAL_BASKETBALL_ASSOCIATION,
    "ABA": League.AMERICAN_BASKETBALL_ASSOCIATION,
    "BAA": League.BASKETBALL_ASSOCIATION_OF_AMERICA,
}


class TeamTotal:
    def __init__(self, team_abbreviation, totals):
        self.team_abbreviation = team_abbreviation
        self.totals = totals

    @property
    def minutes_played(self):
        return self.totals.minutes_played

    @property
    def made_field_goals(self):
        return self.totals.made_field_goals

    @property
    def attempted_field_goals(self):
        return self.totals.attempted_field_goals

    @property
    def made_three_point_field_goals(self):
        return self.totals.made_three_point_field_goals

    @property
    def attempted_three_point_field_goals(self):
        return self.totals.attempted_three_point_field_goals

    @property
    def made_free_throws(self):
        return self.totals.made_free_throws

    @property
    def attempted_free_throws(self):
        return self.totals.attempted_free_throws

    @property
    def offensive_rebounds(self):
        return self.totals.offensive_rebounds

    @property
    def defensive_rebounds(self):
        return self.totals.defensive_rebounds

    @property
    def assists(self):
        return self.totals.assists

    @property
    def steals(self):
        return self.totals.steals

    @property
    def blocks(self):
        return self.totals.blocks

    @property
    def turnovers(self):
        return self.totals.turnovers

    @property
    def personal_fouls(self):
        return self.totals.personal_fouls

    @property
    def points(self):
        return self.totals.points


class PlayerData:
    def __init__(self, name, resource_location, league_abbreviations):
        self.name = name
        self.resource_location = resource_location
        self.league_abbreviations = set(league_abbreviations)



from unittest import TestCase

from basketball_reference_web_scraper.utilities import str_to_int, str_to_float, merge_two_dicts


class TestStrToInt(TestCase):
    def test_empty_string_is_zero(self):
        self.assertEqual(str_to_int(""), 0)

    def test_whitespace_is_zero(self):
        self.assertEqual(str_to_int("    "), 0)

    def test_stringified_number_is_converted(self):
        self.assertEqual(str_to_int("10"), 10)

    def test_stringified_number_with_leading_whitespace_is_converted(self):
        self.assertEqual(str_to_int("  10"), 10)

    def test_stringified_number_with_trailing_whitespace_is_converted(self):
        self.assertEqual(str_to_int("10    "), 10)

    def test_with_default(self):
        self.assertIsNone(str_to_int("", default=None))


class TestStrToFloat(TestCase):
    def test_empty_string_is_zero(self):
        self.assertEqual(str_to_float(""), 0.0)

    def test_whitespace_is_zero(self):
        self.assertEqual(str_to_float("    "), 0.0)

    def test_stringified_number_is_converted(self):
        self.assertEqual(str_to_float("1.234"), 1.234)

    def test_stringified_number_with_leading_whitespace_is_converted(self):
        self.assertEqual(str_to_float("  1.234"), 1.234)

    def test_stringified_number_with_trailing_whitespace_is_converted(self):
        self.assertEqual(str_to_float("1.234    "), 1.234)

    def test_with_default(self):
        self.assertIsNone(str_to_float("", default=None))


class TestMergeTwoDicts(TestCase):
    def test_merges_two_empty_dicts(self):
        self.assertEqual({}, merge_two_dicts({}, {}))

    def test_merges_empty_dict_with_non_empty_dict(self):
        self.assertEqual({"jae": "baebae"}, merge_two_dicts({}, {"jae": "baebae"}))

    def test_merges_non_empty_dict_with_empty_dict(self):
        self.assertEqual({"jae": "baebae"}, merge_two_dicts({"jae": "baebae"}, {}))

    def test_merge_non_empty_dicts_with_unique_keys(self):
        self.assertEqual({"jae": "baebae", "bae": "jadley"}, merge_two_dicts({"jae": "baebae"}, {"bae": "jadley"}))

    def test_merge_non_empty_dicts_with_shared_keys(self):
        self.assertEqual({"jae": "baebae2"}, merge_two_dicts({"jae": "baebae"}, {"jae": "baebae2"}))



from unittest import TestCase, mock
import os

from requests import HTTPError, codes

from basketball_reference_web_scraper.client import player_box_scores
from basketball_reference_web_scraper.data import OutputType, OutputWriteOption
from basketball_reference_web_scraper.errors import InvalidDate


class TestPlayerBoxScores(TestCase):
    def test_get_box_scores(self):
        result = player_box_scores(day=1, month=1, year=2018)
        self.assertIsNotNone(result)

    def test_get_box_scores_for_day_that_does_not_exist(self):
        self.assertRaisesRegex(
            InvalidDate,
            "Date with year set to 2018, month set to 1, and day set to -1 is invalid",
            player_box_scores,
            day=-1, month=1, year=2018)

    def test_get_box_scores_from_2001(self):
        output_file_path = os.path.join(
            os.path.dirname(__file__),
            "./output/2003_10_29_TOR_pbp.csv",
        )
        player_box_scores(
            day=1, month=1, year=2001,
            output_type=OutputType.CSV, output_file_path=output_file_path, output_write_option=OutputWriteOption.WRITE)

    def test_raises_invalid_date_for_nonexistent_dates(self):
        self.assertRaisesRegex(
            InvalidDate,
            "Date with year set to baz, month set to bar, and day set to foo is invalid",
            player_box_scores,
            day="foo", month="bar", year="baz")

    @mock.patch("basketball_reference_web_scraper.client.http_client")
    def test_raises_invalid_date_for_404_response(self, mocked_http_client):
        mocked_http_client.player_box_scores.side_effect = HTTPError(response=mock.Mock(status_code=codes.not_found))
        self.assertRaises(InvalidDate, player_box_scores, day=1, month=1, year=2018)

    @mock.patch("basketball_reference_web_scraper.client.http_client")
    def test_raises_non_404_http_error(self, mocked_http_client):
        mocked_http_client.player_box_scores.side_effect = HTTPError(response=mock.Mock(status_code=codes.server_error))
        self.assertRaises(HTTPError, player_box_scores, day=1, month=1, year=2018)



from unittest import TestCase
from unittest.mock import MagicMock, patch, PropertyMock

from basketball_reference_web_scraper.html import PlayerAdvancedSeasonTotalsRow


class TestPlayerAdvancedSeasonTotalsRow(TestCase):
    def setUp(self):
        self.html = MagicMock()

    def test_position_abbreviations_when_cells_exist(self):
        position_abbreviations = MagicMock()
        text_content = 'some text content'
        position_abbreviations.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[position_abbreviations])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).position_abbreviations, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="pos"]')

    def test_position_abbreviations_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).position_abbreviations, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="pos"]')

    def test_age_when_cells_exist(self):
        age = MagicMock()
        text_content = 'some text content'
        age.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[age])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).age, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="age"]')

    def test_age_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).age, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="age"]')

    def test_team_abbreviation_when_cells_exist(self):
        team_abbreviation = MagicMock()
        text_content = 'some text content'
        team_abbreviation.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[team_abbreviation])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).team_abbreviation, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="team_id"]')

    def test_team_abbreviation_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).team_abbreviation, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="team_id"]')

    def test_games_played_when_cells_exist(self):
        games_played = MagicMock()
        text_content = 'some text content'
        games_played.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[games_played])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).games_played, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="g"]')

    def test_games_played_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).games_played, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="g"]')

    def test_minutes_played_when_cells_exist(self):
        minutes_played = MagicMock()
        text_content = 'some text content'
        minutes_played.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[minutes_played])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).minutes_played, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="mp"]')

    def test_minutes_played_is_empty_string_when_cells_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).minutes_played, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="mp"]')

    def test_player_efficiency_rating_when_cells_exist(self):
        player_efficiency_rating = MagicMock()
        text_content = 'some text content'
        player_efficiency_rating.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[player_efficiency_rating])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).player_efficiency_rating, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="per"]')

    def test_player_efficiency_rating_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).player_efficiency_rating, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="per"]')

    def test_true_shooting_percentage_when_cells_exist(self):
        true_shooting_percentage = MagicMock()
        text_content = 'some text content'
        true_shooting_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[true_shooting_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).true_shooting_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="ts_pct"]')

    def test_true_shooting_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).true_shooting_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="ts_pct"]')

    def test_three_point_attempt_rate_when_cells_exist(self):
        three_point_attempt_rate = MagicMock()
        text_content = 'some text content'
        three_point_attempt_rate.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[three_point_attempt_rate])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).three_point_attempt_rate, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="fg3a_per_fga_pct"]')

    def test_three_point_attempt_rate_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).three_point_attempt_rate, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="fg3a_per_fga_pct"]')

    def test_free_throw_attempt_rate_when_cells_exist(self):
        free_throw_attempt_rate = MagicMock()
        text_content = 'some text content'
        free_throw_attempt_rate.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[free_throw_attempt_rate])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).free_throw_attempt_rate, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="fta_per_fga_pct"]')

    def test_free_throw_attempt_rate_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).free_throw_attempt_rate, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="fta_per_fga_pct"]')

    def test_offensive_rebound_percentage_when_cells_exist(self):
        offensive_rebound_percentage = MagicMock()
        text_content = 'some text content'
        offensive_rebound_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[offensive_rebound_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_rebound_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="orb_pct"]')

    def test_offensive_rebound_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_rebound_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="orb_pct"]')

    def test_defensive_rebound_percentage_when_cells_exist(self):
        defensive_rebound_percentage = MagicMock()
        text_content = 'some text content'
        defensive_rebound_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[defensive_rebound_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_rebound_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="drb_pct"]')

    def test_defensive_rebound_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_rebound_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="drb_pct"]')

    def test_total_rebound_percentage_when_cells_exist(self):
        total_rebound_percentage = MagicMock()
        text_content = 'some text content'
        total_rebound_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[total_rebound_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).total_rebound_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="trb_pct"]')

    def test_total_rebound_percentage_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).total_rebound_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="trb_pct"]')

    def test_assist_percentage_when_cells_exist(self):
        assist_percentage = MagicMock()
        text_content = 'some text content'
        assist_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[assist_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).assist_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="ast_pct"]')

    def test_assist_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).assist_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="ast_pct"]')

    def test_steal_percentage_when_cells_exist(self):
        steal_percentage = MagicMock()
        text_content = 'some text content'
        steal_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[steal_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).steal_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="stl_pct"]')

    def test_steal_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).steal_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="stl_pct"]')

    def test_block_percentage_when_cells_exist(self):
        block_percentage = MagicMock()
        text_content = 'some text content'
        block_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[block_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).block_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="blk_pct"]')

    def test_block_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).block_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="blk_pct"]')

    def test_turnover_percentage_when_cells_exist(self):
        turnover_percentage = MagicMock()
        text_content = 'some text content'
        turnover_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[turnover_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).turnover_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="tov_pct"]')

    def test_turnover_percentage_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).turnover_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="tov_pct"]')

    def test_usage_percentage_when_cells_exist(self):
        usage_percentage = MagicMock()
        text_content = 'some text content'
        usage_percentage.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[usage_percentage])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).usage_percentage, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="usg_pct"]')

    def test_usage_percentage_is_empty_string_when_no_cells_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).usage_percentage, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="usg_pct"]')

    def test_offensive_win_shares_when_cells_exist(self):
        offensive_win_shares = MagicMock()
        text_content = 'some text content'
        offensive_win_shares.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[offensive_win_shares])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_win_shares, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="ows"]')

    def test_offensive_win_shares_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_win_shares, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="ows"]')

    def test_defensive_win_shares_when_cells_exist(self):
        defensive_win_shares = MagicMock()
        text_content = 'some text content'
        defensive_win_shares.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[defensive_win_shares])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_win_shares, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="dws"]')

    def test_defensive_win_shares_is_empty_string_when_no_cells_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_win_shares, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="dws"]')

    def test_win_shares_when_cells_exist(self):
        win_shares = MagicMock()
        text_content = 'some text content'
        win_shares.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[win_shares])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).win_shares, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="ws"]')

    def test_win_shares_is_empty_string_when_no_cells_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).win_shares, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="ws"]')

    def test_win_shares_per_48_minutes_when_cells_exist(self):
        win_shares_per_48_minutes = MagicMock()
        text_content = 'some text content'
        win_shares_per_48_minutes.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[win_shares_per_48_minutes])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).win_shares_per_48_minutes, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="ws_per_48"]')

    def test_win_shares_per_48_minutes_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).win_shares_per_48_minutes, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="ws_per_48"]')

    def test_offensive_plus_minus_when_cells_exist(self):
        offensive_plus_minus = MagicMock()
        text_content = 'some text content'
        offensive_plus_minus.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[offensive_plus_minus])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_plus_minus, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="obpm"]')

    def test_offensive_plus_minus_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).offensive_plus_minus, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="obpm"]')

    def test_defensive_plus_minus_when_cells_exist(self):
        defensive_plus_minus = MagicMock()
        text_content = 'some text content'
        defensive_plus_minus.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[defensive_plus_minus])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_plus_minus, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="dbpm"]')

    def test_defensive_plus_minus_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).defensive_plus_minus, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="dbpm"]')

    def test_plus_minus_when_cells_exist(self):
        plus_minus = MagicMock()
        text_content = 'some text content'
        plus_minus.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[plus_minus])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).plus_minus, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="bpm"]')

    def test_plus_minus_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).plus_minus, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="bpm"]')

    def test_value_over_replacement_player_when_cells_exist(self):
        value_over_replacement_player = MagicMock()
        text_content = 'some text content'
        value_over_replacement_player.text_content = MagicMock(return_value=text_content)
        self.html.xpath = MagicMock(return_value=[value_over_replacement_player])

        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).value_over_replacement_player, text_content)
        self.html.xpath.assert_called_once_with('td[@data-stat="vorp"]')

    def test_value_over_replacement_player_is_empty_string_when_cells_do_not_exist(self):
        self.html.xpath = MagicMock(return_value=[])
        self.assertEqual(PlayerAdvancedSeasonTotalsRow(html=self.html).value_over_replacement_player, '')
        self.html.xpath.assert_called_once_with('td[@data-stat="vorp"]')

    @patch.object(
        PlayerAdvancedSeasonTotalsRow,
        'team_abbreviation',
        new_callable=PropertyMock,
        return_value='Not Total'
    )
    def test_is_not_combined_totals_when_team_abbreviation_is_not_TOT(self, _):
        self.assertFalse(PlayerAdvancedSeasonTotalsRow(html=self.html).is_combined_totals)

    @patch.object(
        PlayerAdvancedSeasonTotalsRow,
        'team_abbreviation',
        new_callable=PropertyMock,
        return_value='TOT'
    )
    def test_is_combined_totals_when_team_abbreviation_is_TOT(self, _):
        self.assertTrue(PlayerAdvancedSeasonTotalsRow(html=self.html).is_combined_totals)



import re


class BasicBoxScoreRow:
    def __init__(self, html):
        self.html = html

    @property
    def playing_time(self):
        cells = self.html.xpath('td[@data-stat="mp"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def minutes_played(self):
        return self.playing_time

    @property
    def made_field_goals(self):
        cells = self.html.xpath('td[@data-stat="fg"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def attempted_field_goals(self):
        cells = self.html.xpath('td[@data-stat="fga"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def made_three_point_field_goals(self):
        cells = self.html.xpath('td[@data-stat="fg3"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def attempted_three_point_field_goals(self):
        cells = self.html.xpath('td[@data-stat="fg3a"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def made_free_throws(self):
        cells = self.html.xpath('td[@data-stat="ft"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def attempted_free_throws(self):
        cells = self.html.xpath('td[@data-stat="fta"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def offensive_rebounds(self):
        cells = self.html.xpath('td[@data-stat="orb"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def defensive_rebounds(self):
        cells = self.html.xpath('td[@data-stat="drb"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def assists(self):
        cells = self.html.xpath('td[@data-stat="ast"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def steals(self):
        cells = self.html.xpath('td[@data-stat="stl"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def blocks(self):
        cells = self.html.xpath('td[@data-stat="blk"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def turnovers(self):
        cells = self.html.xpath('td[@data-stat="tov"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def personal_fouls(self):
        cells = self.html.xpath('td[@data-stat="pf"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def points(self):
        cells = self.html.xpath('td[@data-stat="pts"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''


class PlayerBoxScoreRow(BasicBoxScoreRow):
    def __init__(self, html):
        super().__init__(html=html)

    def __eq__(self, other):
        if isinstance(other, PlayerBoxScoreRow):
            return self.html == other.html
        return False

    @property
    def team_abbreviation(self):
        cells = self.html.xpath('td[@data-stat="team_id"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def location_abbreviation(self):
        cells = self.html.xpath('td[@data-stat="game_location"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def opponent_abbreviation(self):
        cells = self.html.xpath('td[@data-stat="opp_id"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def outcome(self):
        cells = self.html.xpath('td[@data-stat="game_result"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def plus_minus(self):
        cells = self.html.xpath('td[@data-stat="plus_minus"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def game_score(self):
        cells = self.html.xpath('td[@data-stat="game_score"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''


class PlayerIdentificationRow:
    def __init__(self, html):
        self.html = html

    @property
    def player_cell(self):
        cells = self.html.xpath('td[@data-stat="player"]')

        if len(cells) > 0:
            return cells[0]

        return None

    @property
    def slug(self):
        cell = self.player_cell
        if cell is None:
            return ''

        return cell.get('data-append-csv')

    @property
    def name(self):
        cell = self.player_cell
        if cell is None:
            return ''

        return cell.text_content()


class PlayerAdvancedSeasonTotalsTable:
    def __init__(self, html):
        self.html = html

    @property
    def rows_query(self):
        return """
            //table[@id="advanced_stats"]
            /tbody
            /tr[
                contains(@class, "full_table") or 
                contains(@class, "italic_text partial_table") 
                and not(contains(@class, "rowSum"))
            ]
        """

    def get_rows(self, include_combined_totals=False):
        player_advanced_season_totals_rows = []
        for row_html in self.html.xpath(self.rows_query):
            row = PlayerAdvancedSeasonTotalsRow(html=row_html)
            if (include_combined_totals is True and row.is_combined_totals is True) or row.is_combined_totals is False:
                # Basketball Reference includes a "total" row for players that got traded
                # which is essentially a sum of all player team rows
                # I want to avoid including those, so I check the "team" field value for "TOT"
                player_advanced_season_totals_rows.append(row)

        return player_advanced_season_totals_rows


class PlayerSeasonTotalTable:
    def __init__(self, html):
        self.html = html

    @property
    def rows_query(self):
        # Basketball Reference includes individual rows for players that played for multiple teams in a season
        # These rows have a separate class ("italic_text partial_table") than the players that played for a single team
        # across a season.
        return """
            //table[@id="totals_stats"]
            /tbody
            /tr[
                contains(@class, "full_table") or 
                contains(@class, "italic_text partial_table") 
                and not(contains(@class, "rowSum"))
            ]
        """

    @property
    def rows(self):
        player_season_totals_rows = []
        for row_html in self.html.xpath(self.rows_query):
            row = PlayerSeasonTotalsRow(html=row_html)
            # Basketball Reference includes a "total" row for players that got traded
            # which is essentially a sum of all player team rows
            # I want to avoid including those, so I check the "team" field value for "TOT"
            if not row.is_combined_totals:
                player_season_totals_rows.append(row)

        return player_season_totals_rows


class PlayerAdvancedSeasonTotalsRow(PlayerIdentificationRow):
    def __init__(self, html):
        super().__init__(html=html)

    @property
    def position_abbreviations(self):
        cells = self.html.xpath('td[@data-stat="pos"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def age(self):
        cells = self.html.xpath('td[@data-stat="age"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def team_abbreviation(self):
        cells = self.html.xpath('td[@data-stat="team_id"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def games_played(self):
        cells = self.html.xpath('td[@data-stat="g"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def minutes_played(self):
        cells = self.html.xpath('td[@data-stat="mp"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def player_efficiency_rating(self):
        cells = self.html.xpath('td[@data-stat="per"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def true_shooting_percentage(self):
        cells = self.html.xpath('td[@data-stat="ts_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def three_point_attempt_rate(self):
        cells = self.html.xpath('td[@data-stat="fg3a_per_fga_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def free_throw_attempt_rate(self):
        cells = self.html.xpath('td[@data-stat="fta_per_fga_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def offensive_rebound_percentage(self):
        cells = self.html.xpath('td[@data-stat="orb_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def defensive_rebound_percentage(self):
        cells = self.html.xpath('td[@data-stat="drb_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def total_rebound_percentage(self):
        cells = self.html.xpath('td[@data-stat="trb_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def assist_percentage(self):
        cells = self.html.xpath('td[@data-stat="ast_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def steal_percentage(self):
        cells = self.html.xpath('td[@data-stat="stl_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def block_percentage(self):
        cells = self.html.xpath('td[@data-stat="blk_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def turnover_percentage(self):
        cells = self.html.xpath('td[@data-stat="tov_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def usage_percentage(self):
        cells = self.html.xpath('td[@data-stat="usg_pct"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def offensive_win_shares(self):
        cells = self.html.xpath('td[@data-stat="ows"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def defensive_win_shares(self):
        cells = self.html.xpath('td[@data-stat="dws"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def win_shares(self):
        cells = self.html.xpath('td[@data-stat="ws"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def win_shares_per_48_minutes(self):
        cells = self.html.xpath('td[@data-stat="ws_per_48"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def offensive_plus_minus(self):
        cells = self.html.xpath('td[@data-stat="obpm"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def defensive_plus_minus(self):
        cells = self.html.xpath('td[@data-stat="dbpm"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def plus_minus(self):
        cells = self.html.xpath('td[@data-stat="bpm"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def value_over_replacement_player(self):
        cells = self.html.xpath('td[@data-stat="vorp"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def is_combined_totals(self):
        return self.team_abbreviation == "TOT"


class PlayerSeasonTotalsRow(PlayerBoxScoreRow, PlayerIdentificationRow):
    def __init__(self, html):
        super().__init__(html=html)

    @property
    def position_abbreviations(self):
        cells = self.html.xpath('td[@data-stat="pos"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def age(self):
        cells = self.html.xpath('td[@data-stat="age"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def games_played(self):
        cells = self.html.xpath('td[@data-stat="g"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def games_started(self):
        cells = self.html.xpath('td[@data-stat="gs"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def is_combined_totals(self):
        return self.team_abbreviation == "TOT"


class BoxScoresPage:
    def __init__(self, html):
        self.html = html

    @property
    def statistics_tables(self):
        return [
            StatisticsTable(table_html)
            for table_html in self.html.xpath('//table[contains(@class, "stats_table")]')
        ]

    @property
    def basic_statistics_tables(self):
        return [
            table
            for table in self.statistics_tables
            if table.has_basic_statistics is True
        ]


class StatisticsTable:
    def __init__(self, html):
        self.html = html

    @property
    def has_basic_statistics(self):
        return 'game-basic' in self.html.attrib["id"]

    @property
    def team_abbreviation(self):
        # Example id value is box-BOS-game-basic or box-BOS-game-advanced
        match = re.match('^box-(.+)-game', self.html.attrib["id"])
        return match.group(1)

    @property
    def team_totals(self):
        # Team totals are stored as table footers
        footers = self.html.xpath('tfoot/tr')
        if len(footers) > 0:
            return BasicBoxScoreRow(html=footers[0])

        return None


class DailyLeadersPage:
    def __init__(self, html):
        self.html = html

    @property
    def daily_leaders(self):
        return [
            PlayerGameBoxScoreRow(row_html)
            for row_html in self.html.xpath('//table[@id="stats"]//tbody/tr[not(contains(@class, "thead"))]')
        ]


class PlayerSeasonBoxScoresPage:
    def __init__(self, html):
        self.html = html

    @property
    def regular_season_box_scores_table_query(self):
        return '//table[@id="pgl_basic"]'

    @property
    def regular_season_box_scores_table(self):
        matching_tables = self.html.xpath(self.regular_season_box_scores_table_query)

        if len(matching_tables) != 1:
            return None

        return RegularSeasonPlayerBoxScoresTable(html=matching_tables[0])


class PlayerSeasonBoxScoresTable:
    def __init__(self, html):
        self.html = html

    @property
    def rows_query(self):
        raise NotImplementedError()

    @property
    def rows(self):
        return [
            PlayerSeasonBoxScoresRow(html=row_html)
            for row_html in self.html.xpath(self.rows_query)
        ]


class RegularSeasonPlayerBoxScoresTable(PlayerSeasonBoxScoresTable):
    @property
    def rows_query(self):
        # Every 20 rows, there's a row that has the column header values - those should be ignored
        return '//tbody' \
               '/tr[not(contains(@class, "thead"))]'


class PlayerSeasonBoxScoresRow(PlayerBoxScoreRow):
    def __init__(self, html):
        super().__init__(html)

    def __eq__(self, other):
        if isinstance(other, PlayerSeasonBoxScoresRow):
            return self.html == other.html
        return False

    @property
    def is_active(self):
        # When a player is not active (for a reason like "Inactive", "Did Not Play", "Did Not Dress")
        # the game played counter is blank (and a "reason" column will exist)
        cells = self.html.xpath('td[@data-stat="reason"]')
        return len(cells) < 1

    @property
    def date(self):
        cells = self.html.xpath('td[@data-stat="date_game"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def points_scored(self):
        cells = self.html.xpath('td[@data-stat="pts"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''


class PlayerGameBoxScoreRow(PlayerBoxScoreRow, PlayerIdentificationRow):
    def __init__(self, html):
        super().__init__(html)


class PlayByPlayPage:
    def __init__(self, html):
        self.html = html

    @property
    def table_query(self):
        return '//table[@id="pbp"]'

    @property
    def team_names_query(self):
        return \
            '//*[@id="content"]' \
            '//div[@class="scorebox"]' \
            '//div[@itemprop="performer"]' \
            '//a[@itemprop="name"]'

    @property
    def play_by_play_table(self):
        return PlayByPlayTable(html=self.html.xpath(self.table_query)[0])

    @property
    def team_names(self):
        names = self.html.xpath(self.team_names_query)

        return [
            name.text_content()
            for name in names
        ]

    @property
    def away_team_name(self):
        return self.team_names[0]

    @property
    def home_team_name(self):
        return self.team_names[1]


class PlayByPlayTable:
    def __init__(self, html):
        self.html = html

    @property
    def rows(self):
        return [
            PlayByPlayRow(html=row_html)
            # Ignore first row in table
            for row_html in self.html[1:]
        ]


class PlayByPlayRow:
    def __init__(self, html):
        self.html = html

    @property
    def timestamp_cell(self):
        return self.html[0]

    @property
    def timestamp(self):
        return self.timestamp_cell.text_content().strip()

    @property
    def away_team_play_description(self):
        return self.html[1].text_content().strip()

    @property
    def home_team_play_description(self):
        return self.html[5].text_content().strip()

    @property
    def is_away_team_play(self):
        return self.away_team_play_description != ""

    @property
    def is_home_team_play(self):
        return self.home_team_play_description != ""

    @property
    def formatted_scores(self):
        return self.html[3].text_content().strip()

    @property
    def is_start_of_period(self):
        return self.timestamp_cell.get('colspan') == '6'

    @property
    def has_play_by_play_data(self):
        # TODO: @jaebradley refactor this to be slightly clearer
        # Need to avoid rows that indicate start of period
        # Or denote tipoff / end of period (colspan = 5)
        # Or are one of the table headers for each period group (aria-label = Time)
        return not self.is_start_of_period \
               and self.html[1].get('colspan') != '5' \
               and self.timestamp_cell.get('aria-label') != 'Time'


class DailyBoxScoresPage:
    def __init__(self, html):
        self.html = html

    @property
    def game_url_paths_query(self):
        return '//td[contains(@class, "gamelink")]/a'

    @property
    def game_url_paths(self):
        game_links = self.html.xpath(self.game_url_paths_query)
        return [game_link.attrib['href'] for game_link in game_links]


class SchedulePage:
    def __init__(self, html):
        self.html = html

    @property
    def other_months_schedule_links_query(self):
        return '//div[@id="content"]' \
               '/div[@class="filter"]' \
               '/div[not(contains(@class, "current"))]' \
               '/a'

    @property
    def rows_query(self):
        return '//table[@id="schedule"]//tbody/tr'

    @property
    def other_months_schedule_urls(self):
        links = self.html.xpath(self.other_months_schedule_links_query)
        return [
            link.attrib['href']
            for link in links
        ]

    @property
    def rows(self):
        return [
            ScheduleRow(html=row)
            for row in self.html.xpath(self.rows_query)
            # Every row in each month's schedule table represents a game
            # except for the row where the only content is "Playoffs"
            if row.text_content() != 'Playoffs'
        ]


class ScheduleRow:
    def __init__(self, html):
        self.html = html

    def __eq__(self, other):
        if isinstance(other, ScheduleRow):
            return self.html == other.html
        return False

    @property
    def start_date(self):
        cells = self.html.xpath('th[@data-stat="date_game"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def start_time_of_day(self):
        cells = self.html.xpath('td[@data-stat="game_start_time"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def away_team_name(self):
        cells = self.html.xpath('td[@data-stat="visitor_team_name"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def home_team_name(self):
        cells = self.html.xpath('td[@data-stat="home_team_name"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def away_team_score(self):
        cells = self.html.xpath('td[@data-stat="visitor_pts"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''

    @property
    def home_team_score(self):
        cells = self.html.xpath('td[@data-stat="home_pts"]')

        if len(cells) > 0:
            return cells[0].text_content()

        return ''


class SearchPage:
    def __init__(self, html):
        self.html = html

    @property
    def nba_aba_baa_players_content_query(self):
        return '//div[@id="searches"]/div[@id="players"]'

    @property
    def nba_aba_baa_players_pagination_links_query(self):
        return '{NBA_ABA_PLAYERS_CONTENT_QUERY}/div[@class="search-pagination"]/a'.format(
            NBA_ABA_PLAYERS_CONTENT_QUERY=self.nba_aba_baa_players_content_query
        )

    @property
    def nba_aba_baa_player_search_items_query(self):
        return '{NBA_ABA_PLAYERS_CONTENT_QUERY}/div[@class="search-item"]'.format(
            NBA_ABA_PLAYERS_CONTENT_QUERY=self.nba_aba_baa_players_content_query
        )

    @property
    def nba_aba_baa_players_pagination_links(self):
        return self.html.xpath(self.nba_aba_baa_players_pagination_links_query)

    @property
    def nba_aba_baa_players_pagination_url(self):
        links = self.nba_aba_baa_players_pagination_links

        if len(links) <= 0:
            return None

        first_link = links[0]

        if len(links) == 1:
            if first_link.text_content() == 'Previous 100 Results':
                return None

            return first_link.attrib["href"]

        return links[1].attrib["href"]

    @property
    def nba_aba_baa_players(self):
        return [
            PlayerSearchResult(html=result_html)
            for result_html in self.html.xpath(self.nba_aba_baa_player_search_items_query)
        ]


class SearchResult:
    def __init__(self, html):
        self.html = html

    @property
    def resource_link_query(self):
        return './div[@class="search-item-name"]//a'

    @property
    def resource_link(self):
        links = self.html.xpath(self.resource_link_query)

        if len(links) > 0:
            return links[0]

        return None

    @property
    def resource_location(self):
        link = self.resource_link

        if link is None:
            return None

        return link.attrib["href"]

    @property
    def resource_name(self):
        link = self.resource_link

        if link is None:
            return None

        return link.text_content()

    def __eq__(self, other):
        if isinstance(other, SearchResult):
            return self.html == other.html
        return False


class PlayerSearchResult(SearchResult):
    @property
    def league_abbreviation_query(self):
        return './div[@class="search-item-league"]'

    @property
    def league_abbreviations(self):
        abbreviations = self.html.xpath(self.league_abbreviation_query)
        
        if len(abbreviations) > 0:
            return abbreviations[0].text_content()

        return None
    

class PlayerPageTotalsRow:
    def __init__(self, html):
        self.html = html
    
    @property
    def league_abbreviation(self):
        league_abbreviation_cells = self.html.xpath('.//td[@data-stat="lg_id"]')

        if len(league_abbreviation_cells) > 0:
            return league_abbreviation_cells[0].text_content()

        return None

    def __eq__(self, other):
        if isinstance(other, PlayerPageTotalsRow):
            return self.html == other.html
        return False


class PlayerPageTotalsTable:
    def __init__(self, html):
        self.html = html

    @property
    def rows(self):
        return [
            PlayerPageTotalsRow(html=row_html)
            for row_html in self.html.xpath('.//tbody/tr')
        ]

    def __eq__(self, other):
        if isinstance(other, PlayerPageTotalsTable):
            return self.html == other.html
        return False


class PlayerPage:
    def __init__(self, html):
        self.html = html
    
    @property
    def name(self):
        name_headers = self.html.xpath('.//h1[@itemprop="name"]')

        if len(name_headers) > 0:
            return name_headers[0].text_content()

        return None

    @property
    def totals_table(self):
        totals_tables = self.html.xpath('.//table[@id="per_game"]')

        if len(totals_tables) > 0:
            return PlayerPageTotalsTable(html=totals_tables[0])

        return None



from unittest import TestCase
from unittest.mock import MagicMock

from basketball_reference_web_scraper.html import PlayerSeasonBoxScoresTable


class TestPlayerSeasonBoxScoresTable(TestCase):
    def setUp(self):
        self.html = MagicMock()

    def test_rows_query_raises_not_implemented_error(self):
        table = PlayerSeasonBoxScoresTable(html=self.html)
        self.assertRaises(
            NotImplementedError,
            lambda: table.rows_query,
        )

    def test_rows_raises_not_implemented_error_when_rows_query_is_not_overridden(self):
        table = PlayerSeasonBoxScoresTable(html=self.html)
        self.assertRaises(
            NotImplementedError,
            lambda: table.rows,
        )




import json
import os
from datetime import date, datetime
from pathlib import Path
from unittest import TestCase

import pytz

from basketball_reference_web_scraper.client import season_schedule
from basketball_reference_web_scraper.data import OutputType, Team
from basketball_reference_web_scraper.errors import InvalidSeason


class TestSeasonScheduleInMemoryOutput(TestCase):
    def test_2018_season_schedule_length(self):
        result = season_schedule(season_end_year=2018)
        self.assertEqual(1312, len(result))

    def test_first_game_of_2018_season(self):
        result = season_schedule(season_end_year=2018)
        self.assertEqual(
            result[0],
            {
                "away_team": Team.BOSTON_CELTICS,
                "away_team_score": 99,
                "home_team": Team.CLEVELAND_CAVALIERS,
                "home_team_score": 102,
                "start_time": datetime(2017, 10, 18, 0, 1, tzinfo=pytz.utc),
            },
        )

    def test_last_game_of_2018_season(self):
        result = season_schedule(season_end_year=2018)
        self.assertEqual(
            result[1311],
            {
                "away_team": Team.GOLDEN_STATE_WARRIORS,
                "away_team_score": 108,
                "home_team": Team.CLEVELAND_CAVALIERS,
                "home_team_score": 85,
                "start_time": datetime(2018, 6, 9, 1, 0, tzinfo=pytz.utc)
            }
        )

    def test_future_season_schedule_throws_invalid_season_error(self):
        current_year = date.today().year
        future_year = current_year + 10
        expected_message = "Season end year of {future_year} is invalid".format(future_year=future_year)
        self.assertRaisesRegex(
            InvalidSeason,
            expected_message,
            season_schedule,
            season_end_year=future_year
        )


class TestSeasonScheduleCSVOutput(TestCase):
    def setUp(self):
        self.output_2018_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2018_season_schedule.csv"
        )
        self.expected_output_2018_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2018_season_schedule.csv"
        )
        self.output_2001_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2001_season_schedule.csv"
        )
        self.expected_output_2001_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2001_season_schedule.csv"
        )

    def tearDown(self):
        if Path(self.output_2018_file_path).exists():
            os.remove(self.output_2018_file_path)

        if Path(self.output_2001_file_path).exists():
            os.remove(self.output_2001_file_path)

    def test_2018_season_schedule_csv(self):
        season_schedule(season_end_year=2018, output_type=OutputType.CSV, output_file_path=self.output_2018_file_path)
        with open(self.output_2018_file_path, "r") as output_file, \
                open(self.expected_output_2018_file_path, "r") as expected_output_file:
            self.assertEqual(
                output_file.readlines(),
                expected_output_file.readlines()
            )

    def test_2017_season_schedule_csv(self):
        season_schedule(season_end_year=2001, output_type=OutputType.CSV, output_file_path=self.output_2001_file_path)
        with open(self.output_2001_file_path, "r") as output_file, \
                open(self.expected_output_2001_file_path, "r") as expected_output_file:
            self.assertEqual(
                output_file.readlines(),
                expected_output_file.readlines()
            )


class TestSeasonScheduleJSONOutput(TestCase):
    def setUp(self):
        self.output_2018_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2018_season_schedule.json"
        )
        self.expected_output_2018_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2018_season_schedule.json"
        )
        self.output_2001_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/2001_season_schedule.json"
        )
        self.expected_output_2001_file_path = os.path.join(
            os.path.dirname(__file__),
            "../output/expected/2001_season_schedule.json"
        )

    def tearDown(self):
        if Path(self.output_2018_file_path).exists():
            os.remove(self.output_2018_file_path)

        if Path(self.output_2001_file_path).exists():
            os.remove(self.output_2001_file_path)

    def test_2018_season_schedule_in_memory_json(self):
        result = season_schedule(season_end_year=2018, output_type=OutputType.JSON)
        with open(self.expected_output_2018_file_path, "r") as expected_output_file:
            self.assertEqual(
                json.loads(result),
                json.load(expected_output_file)
            )

    def test_writing_2018_season_schedule_json_file(self):
        season_schedule(season_end_year=2018, output_type=OutputType.JSON, output_file_path=self.output_2018_file_path)
        with open(self.output_2018_file_path, "r") as output_file, \
                open(self.expected_output_2018_file_path, "r") as expected_output_file:
            self.assertEqual(
                json.load(output_file),
                json.load(expected_output_file),
            )

    def test_writing_2017_season_schedule_json_file(self):
        season_schedule(season_end_year=2001, output_type=OutputType.JSON, output_file_path=self.output_2001_file_path)
        with open(self.output_2001_file_path, "r") as output_file, \
                open(self.expected_output_2001_file_path, "r") as expected_output_file:
            self.assertEqual(
                json.load(output_file),
                json.load(expected_output_file),
            )
